#ifndef ASCIIDATAFILE_H_
#define ASCIIDATAFILE_H_

#include <DataFile.h>
#include <fstream>

/**
 * Stores the data in ASCII (binary) mode
 */
class AsciiDataFile : public DataFile
{
    private:
        // output file stream
        std::ofstream *ofile;

        // Pointer to the end of the header
        std::ofstream::pos_type end_of_header;
    protected:
        static const unsigned int data_header;
        static const unsigned int header_length;

    public:
        // Constructor
        AsciiDataFile(const std::string &name);
        ~AsciiDataFile();

        // Opens the file
        virtual void open(const std::string &nam="");
        // Closes the file
        virtual void close();

        // Writes a user defined buffer
        virtual void write(BlockType type, size_t size, const char *buff);

        // Writes the current event
        virtual void write(const AlibavaData *evt);

        /** Writes the current value of pedestals and noise
         *  Also stores the run type, given in type
         */
        virtual void write(int type, double *p, double *n);

};

#endif /*ASCIIDATAFILE_H_*/
