/*
 * GtkUtils.cc
 *
 *  Created on: Oct 29, 2009
 *      Author: lacasta
 */

#include <iostream>
#include <string>
#include <sstream>
#include <cstdio>
#include <cstring>
#include <vector>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <glib.h>
#include "GtkUtils.h"
#include "string_utils.h"
#include "utils.h"
#include "path_utils.h"

/**
 * Filters are given as an array of filters in the form <name>|<regex>
 */

#ifdef G_OS_WIN32
#include <windows.h>
#include <Shlobj.h>
std::string get_folder_name(const char *title, const char *init_folder)
{
    std::string out;
    BROWSEINFO bi = { 0 };
    bi.lpszTitle = title;
    LPITEMIDLIST pidl = SHBrowseForFolder ( &bi );
    if ( pidl != 0 )
    {
        // get the name of the folder
        TCHAR path[MAX_PATH];
        if ( SHGetPathFromIDList ( pidl, path ) )
        {
            out = path;
        }

        // free memory used
        IMalloc * imalloc = 0;
        if ( SUCCEEDED( SHGetMalloc ( &imalloc )) )
        {
            imalloc->Free ( pidl );
            imalloc->Release ( );
        }
    }
    return out;
}

std::string get_file_name(const char *title, GetFileNameAction for_open, const char *init_folder, const char **filters )
{
    OPENFILENAME of;
    int retval;
    char fname[512];
    char startdir[512];
    fname[0] = '\0';

    if ( for_open == a_OpenFolder )
        return get_folder_name(title, init_folder);

    GString *str = g_string_new("");
    if (filters)
    {
        int i, nchar;
        for (i = 0; filters[i]; i++)
        {
        	if ( !strcmp(filters[i], "All|*") )
        		continue;

            gchar *ff = g_strdup(filters[i]);
            gchar *p = strchr(ff,'|');
            if (*p)
            {
                *p++ = 0;
                g_string_append(str, ff);
                g_string_append_c(str, '\0');
                g_string_append(str, p);
                g_string_append_c(str, '\0');
            }
            else
            {
                g_string_append(str, "My filter");
                g_string_append_c(str, '\0');
                g_string_append(str, ff);
                g_string_append_c(str, '\0');
            }
            g_free(ff);
        }
        g_string_append(str, "All");
        g_string_append_c(str, '\0');
        g_string_append(str, "*.*");
        g_string_append_c(str, '\0');
        g_string_append_c(str, '\0');
    }


    getcwd(startdir, 512);
    /* initialize file dialog info struct */
    memset(&of, 0, sizeof of);
#ifdef OPENFILENAME_SIZE_VERSION_400
    of.lStructSize = OPENFILENAME_SIZE_VERSION_400;
#else
    of.lStructSize = sizeof of;
#endif
    of.hwndOwner = NULL;
    of.lpstrFilter = str->str;
    of.lpstrCustomFilter = 0;
    of.nFilterIndex = 1;
    of.lpstrFile = fname;
    of.nMaxFile = sizeof fname;
    of.lpstrFileTitle = NULL;
    of.lpstrInitialDir = init_folder;
    of.lpstrTitle = title;
    of.lpstrDefExt = NULL;
    of.Flags = OFN_HIDEREADONLY;

    if (for_open == a_OpenFile)
        retval = GetOpenFileName(&of);
    else  /* a file save action */
        retval = GetSaveFileName(&of);

    g_string_free(str, TRUE);

    if (!retval) {
        return std::string();
    }
    return fname;
}



gchar *get_registry_key(const gchar *path, const gchar *key);

gchar *get_help_path()
{
    gchar *path = get_registry_key("SOFTWARE\\AlibavaSL", "HELPDIR");
}


#else
std::string get_file_name(const char *title, GetFileNameAction for_open, const char *init_folder=0, const char **filters)
{
    std::string out;
    GtkWidget *fsel;
#if  GTK_CHECK_VERSION(2,4,0)
    int i;
    GtkFileFilter *filter = 0;

    GtkFileChooserAction action;
    switch (for_open)
    {
        case a_OpenFile:
            action = GTK_FILE_CHOOSER_ACTION_OPEN;
            break;
        case a_SaveFile:
            action = GTK_FILE_CHOOSER_ACTION_SAVE;
            break;
        case a_OpenFolder:
            action = GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER;
            break;
    }

    fsel = gtk_file_chooser_dialog_new(title, NULL, action,
                                       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                       GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
                                       NULL);
    if (init_folder)
        gtk_file_chooser_set_current_folder( GTK_FILE_CHOOSER(fsel), init_folder);

    if (filters)
    {
        for (i = 0; filters[i]; i++)
        {
            filter = gtk_file_filter_new();
            gchar *ff = g_strdup(filters[i]);
            gchar *p = strchr(ff,'|');
            if (!p)
                gtk_file_filter_add_pattern(filter, filters[i]);
            else
            {
                char *q=0, *w;
                *p++ = 0;
                gtk_file_filter_set_name(filter, ff);
                w = p;
                do
                {
                    q = strchr(w,',');
                    if (q)
                        *q='\0';

                    gtk_file_filter_add_pattern(filter, w);
                    if (q)
                        w = ++q;

                } while (q);

            }
            g_free(ff);
            gtk_file_chooser_add_filter( GTK_FILE_CHOOSER (fsel), filter);
        }
    }


#define fsel_get_file(fsel)    gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (fsel))
#define the_good_answer  GTK_RESPONSE_ACCEPT
#else
    fsel = gtk_file_selection_new(title);
#define fsel_get_file(fsel)    gtk_file_selection_get_filename (GTK_FILE_SELECTION (fsel))
#define the_good_answer  GTK_RESPONSE_OK
#endif

    if (gtk_dialog_run(GTK_DIALOG(fsel)) == the_good_answer)
        out = fsel_get_file(fsel);

    gtk_widget_destroy(fsel);
    return out;
}
#endif


void
gtk_help_display(const gchar *link_id)
{
    GError *err=0;
#ifndef G_OS_WIN32
    gchar *path;
    gchar *help_dir = getenv("ALIBAVA_HELPDIR");
    if (help_dir)
        path = g_strdup( help_dir );
    else
        path = g_build_filename(HELPDIR, "alibava.html", NULL);

#else
    gchar *gdir = get_help_path();
    gchar *path = g_build_filename(gdir, "alibava.html", NULL);
    g_free(gdir);
#endif
    gchar *uri = g_filename_to_uri(path, 0, &err);
    if (err)
    {
        err_msg(_("Error opening help file: %s\n"), err->message);
        g_error_free(err);
        err = 0;
    }
    else
    {
    /** TODO: what do we do if gtk_show_uri is not there in older Gtk version ?
     * 
     */
#if  GTK_CHECK_VERSION(2,14,0)
#ifdef G_OS_WIN32
        /* This is a ugly hack, should be fixed in glib */
        if (!(ShellExecute(NULL, "open", uri, NULL, NULL, SW_SHOWNORMAL) > (HINSTANCE)32)) {
            err = g_error_new_literal(0, 0, "failed to open webpage");
        }
#else
#ifdef __APPLE__
        std::ostringstream ostr;
        ostr << "open " << uri;
        g_spawn_command_line_async (ostr.str().c_str(), &err);
#else
        gtk_show_uri(0, uri, GDK_CURRENT_TIME, &err);
#endif
#endif
        if (err)
        {
            err_msg(_("Error opening help file: %s\n"), err->message);
            err_msg(_("\t%s"), uri);
            g_error_free(err);
        }
#endif
    }
    g_free(uri);
    g_free(path);
}


std::string get_a_string(const std::string &title, const std::string &prompt)
{
    std::string out;
    GtkWidget *dialog = gtk_dialog_new_with_buttons (
                            title.c_str(),
                            (GtkWindow *)0,
                            (GtkDialogFlags)(GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT),
                            GTK_STOCK_OK,
                            GTK_RESPONSE_ACCEPT,
                            GTK_STOCK_CANCEL,
                            GTK_RESPONSE_REJECT,
                            NULL);

    GtkWidget *vbox = gtk_vbox_new( FALSE, 2 );
    gtk_container_set_border_width( GTK_CONTAINER(vbox), 5 );
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), vbox, FALSE, FALSE, 0);

    GtkWidget *label = gtk_label_new(prompt.c_str());
    gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
    gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

    GtkWidget *answ = gtk_entry_new();
    gtk_box_pack_start(GTK_BOX(vbox), answ, FALSE, FALSE, 0);

    gtk_button_box_set_layout( GTK_BUTTON_BOX(GTK_DIALOG(dialog)->action_area),
                               GTK_BUTTONBOX_END);

    gtk_widget_show_all( dialog );

    if ( gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT )
    {
        out = gtk_entry_get_text(GTK_ENTRY(answ));

    }
    gtk_widget_hide(dialog);
    gtk_widget_destroy(dialog);

    return out;
}

int get_a_string_list(const std::string &title, const std::string &prompt, std::vector<std::string> &values)
{
    std::vector<std::string> msg;
    stringtok< std::vector<std::string> >(msg, prompt,":;");
    std::vector<std::string>::iterator ip;

    values.clear();
    GtkWidget *dialog = gtk_dialog_new_with_buttons (
                                                     title.c_str(),
                                                     (GtkWindow *)0,
                                                     (GtkDialogFlags)(GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT),
                                                     GTK_STOCK_OK,
                                                     GTK_RESPONSE_ACCEPT,
                                                     GTK_STOCK_CANCEL,
                                                     GTK_RESPONSE_REJECT,
                                                     NULL);

    GtkWidget *vbox = gtk_vbox_new( FALSE, 2 );
    gtk_container_set_border_width( GTK_CONTAINER(vbox), 5 );
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), vbox, FALSE, FALSE, 0);

    std::vector< GtkWidget *> entries;
    for (ip=msg.begin();ip!=msg.end();ip++)
    {
        GtkWidget *label = gtk_label_new((*ip).c_str());
        gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
        gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

        GtkWidget *answ = gtk_entry_new();
        gtk_box_pack_start(GTK_BOX(vbox), answ, FALSE, FALSE, 0);
        entries.push_back(answ);
    }
    gtk_button_box_set_layout( GTK_BUTTON_BOX(GTK_DIALOG(dialog)->action_area),
                               GTK_BUTTONBOX_END);

    gtk_widget_show_all( dialog );

    if ( gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT )
    {
        std::vector< GtkWidget *>::iterator iw;
        for (iw=entries.begin(); iw!=entries.end();iw++)
        {
            values.push_back( gtk_entry_get_text(GTK_ENTRY( *iw )));
        }

    }
    gtk_widget_hide(dialog);
    gtk_widget_destroy(dialog);
    return values.size();
}


/* This is an internally used function to create pixmaps. */
GdkPixbuf* create_pixbuf(const gchar *filename)
{
    GdkPixbuf *pixbuf;
    GError *error = NULL;

    if (!filename || !filename[0])
        return NULL;

    if (!filename)
    {
        err_msg( _("Couldn't find pixmap file: %s"), filename);
        return NULL;
    }

    pixbuf = gdk_pixbuf_new_from_file(filename, &error);
    if (!pixbuf)
    {
        err_msg(_("Failed to load pixbuf file: %s: %s\n"), filename, error->message);
        g_error_free(error);
    }
    return pixbuf;
}


int
select_text_option(const std::string &title, const std::string &prompt, std::vector<std::string> &values, int start)
{
    int rc = -1;
    GtkWidget *dlg;
    GtkWidget *vbox;
    GtkWidget *label;
    GtkWidget *omenu;
    GtkBox *area;
    std::ostringstream ostr;

    dlg = gtk_dialog_new_with_buttons(
            title.c_str(),
            NULL,
            (GtkDialogFlags)(GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT),
            GTK_STOCK_OK,
            GTK_RESPONSE_ACCEPT,
            GTK_STOCK_CANCEL,
            GTK_RESPONSE_REJECT,
            NULL
    );

    area = GTK_BOX( gtk_dialog_get_content_area( GTK_DIALOG(dlg) ) );
    vbox = gtk_vbox_new(FALSE,10);
    gtk_box_pack_start(area, vbox, TRUE, TRUE,0);

    ostr << "<b><big>" << prompt << "</big></b>";
    label = gtk_label_new(0);
    gtk_label_set_markup(GTK_LABEL(label), ostr.str().c_str());
    gtk_misc_set_alignment(GTK_MISC(label),0,0.5);
    gtk_box_pack_start(GTK_BOX(vbox),label,TRUE,TRUE,0);


#if  GTK_CHECK_VERSION(2,24,0)
    omenu = gtk_combo_box_text_new();
#else
    omenu = gtk_combo_box_entry_new_text();
#endif
    std::vector<std::string>::iterator ip;
    for (ip=values.begin(); ip!=values.end(); ++ip)
    {
#if  GTK_CHECK_VERSION(2,24,0)
        gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (omenu), ip->c_str());
#else
        gtk_combo_box_append_text( GTK_COMBO_BOX(omenu), ip->c_str() );
#endif
    }
    gtk_combo_box_set_active(GTK_COMBO_BOX(omenu), start);
    gtk_box_pack_start(GTK_BOX(vbox), omenu, FALSE, TRUE, 0);


    gtk_widget_show_all(dlg);

    if ( gtk_dialog_run( GTK_DIALOG(dlg) )==GTK_RESPONSE_ACCEPT )
    {
        rc = gtk_combo_box_get_active(GTK_COMBO_BOX(omenu));
    }

    gtk_widget_hide(dlg);
    gtk_widget_destroy(dlg);
    return rc;
}


