/*
 * HDFDataFile.h
 *
 *  Created on: Feb 27, 2014
 *      Author: lacasta
 */

#ifndef HDFDATAFILE_H_
#define HDFDATAFILE_H_

#include <DataFile.h>

/**
 * Stores the data using the HDF library
 */
struct HDFPrivate;


/**
 * class HDFDataFile.
 *
 * A DataFile object that uses the HDF5 file format
 */
class HDFDataFile : public DataFile
{
    private:
        HDFPrivate *priv;

    public:
        HDFDataFile(const std::string &name);
        virtual ~HDFDataFile();

        // Opens the file
        void open(const std::string &nam="");
        // Closes the file
        void close();

        /**
         * Describes the UserData format. It follows the idea of the
         * struct module in Python. The format is described as follows
         *
         * A comma separated list of items of the form
         *
         * name/fmt
         *
         * where
         *  name is the name to given to the array and
         *  fmt  a string describing the type. It follows the convention
         *       of the struct module in Python
         *
         * Type description:
         *
         * Optional first char
         *   =: native order, std. size & alignment
         *   <: little-endian, std. size & alignment
         *   >: big-endian, std. size & alignment
         *
         * Followed by a number to specify dimension
         * Then comes the type:
         *    c     - char
         *    b/B   - signed/unsigned byte
         *    h/H   - signed/unsigned short
         *    i/I   - signed/unsigned int
         *    l/L   - signed/unsigned long
         *    f     - float
         *    d     - double
         *
         * No other types are allowed
         */
        void set_data_format(const char *fmt);

        /**
         * Writes a user defined buffer
         */
        void write(BlockType type, size_t size, const char *buff);

        // Writes the current event
        void write(const AlibavaData *evt);

        /** Writes the current value of pedestals and noise
         *  Also stores the run type, given in type
         */
        void write(int type, double *p, double *n);

        /**
         * Receives the scan definition
         */
        void scan_definition(int type, int npoints, int from, int to, int nevts);

        /// Receives the setup description
        void describe_setup(int run_type, int nchip, int chip_mask, int firmware);

        /// Receives the information about a just finished scan point
        void new_point(int start, int end, double value);

        /// Flushes to disk
        void flush();
};

#endif /* HDFDATAFILE_H_ */
