        //#define DEBUG 1
#include <glib.h>
#ifdef HAVE_CONFIG_H
#ifdef G_OS_WIN32
#include "config_win.h"
#else
#include <config.h>
#endif
#else
#warning "NO config"
#endif

#define VERBOSE(fmt, args...) \
    if ( verbose() ) { \
        printf(fmt , ## args); \
        fflush(stdout); \
    }


#include <iostream>
#include <iomanip>
#include <sstream>
#include <cmath>
#include <ctime>
#include <map>
#include <glib.h>
#include <unistd.h>
#include "utils.h"
#include "Machine.h"
#include "Alibava.h"


static double randuniform();
static double randgauss(double mean, double stdev);
static double landau(double mean, double stdev);
static std::map<int, std::string> alibava_states;
static time_t clock_0 = 0;

const double time_offset = 10;
static double tau1(double Vfp)
{
    return 100.0 + Vfp*( 1.5 + 0.102*Vfp );
}

static
double beetle(double *x, double *par)
{
   double t = (*x + par[0]);
   double tau1 = par[1];
   double tau2 = par[2];
   double K = par[3];
   double value = K*(tau1/(tau1-tau2))*(exp(-t/tau1)-exp(-t/tau2));
//   printf("t %f (*x) %f F=%f\n", t, *x, value);

   return value;
}

static double beetle_max(double *par, double *xval=0)
{
   double K0 = par[3];
   double tau1 = par[1];
   double tau2 = par[2];
   double xx   = (1/tau1 - 1/tau2);
   double vmax, xmax = log(tau2/tau1)/xx - par[0];
   vmax =   beetle(&xmax, par);
   if (xval)
       *xval = xmax;

   return vmax;
}

static unsigned short comparator_offset[256] =
{
 124,130,137,116,126,132,118,130,129,133,122,144,140,123,124,125,
 122,137,117,120,115,129,137,115,127,128,130,122,133, 99,114,114,
 131,108,151,134,146,140,115,125,115,135,136,114,128,107,132,121,
 142,131,138,118,130,123,133,124,135,146,132,129,121,138,122,123,
 125,126,124,130,129,143,119,126,133,113,119,104,141,142,154,120,
 114,130,125,136,122,120,133,127,124,128,125,118,143,133,121,129,
 130,127,128,130,134,133,134,125,147,124,127,130,118,121,134,132,
 120,121,154,132,133,125,132,140,132,118,134,124,118,135,142,132,
 128,125,125,121,124,143,123,129,134,117,128,130,131,133,131,132,
 138,120,126,127,131,141,138,118,130,140,141,130,126,130,128,133,
 143,139,127,112,130,124,125,133,129,128,137,141,121,116,106,127,
 126,140,120,127,135,113,121,133,135,135,118,116,136,132,143,156,
 135,121,123,115,130,119,141,148,135,119,136,131,124,119,127,125,
 138,126,127,113,142,132,148,123,123,130,131,122,140,123,124,106,
 126,126,128,122,134,124,118,124,125,115,130,122,130,131,119,130,
 158,139,127,136,123,122,131,121,140,107,122,127,123,135,124,120
};

template<class T>
T reverse_bit(T v)
{
   T r = v; // r will be reversed bits of v; first get LSB of v
   int s = sizeof(v) * CHAR_BIT - 1; // extra shift needed at end

   for (v >>= 1; v; v >>= 1)
   {
      r <<= 1;
      r |= v & 1;
      s--;
   }
   r <<= s; // shift when v's highest bits are zero
   return r;
}

const char *TheMachine::sstate[] =
{ "NotInitialized", "SoftwareReset", "WaitingForHandshake", "Waiting", "Initialized", "ReadyForData",
        "BeetleConfiguration", "BeetleConfigReady", "BeetleCalibration",
        "AlibavaConf","AlibavaconfReady", "AlibavaConfEnd",
        "TriggerCount", "TriggerCountReady", "TriggerCountEnd",
        "TriggerIn", "TriggerDACs", "LaserSynchronization",
        "PrepareAcquisition", "Acquiring", "ReadingData", "WaitReadAck", 0 };

TheMachine::TheMachine() :
   fw_version(1), chip_mask(3), nreset(4),
   _state(0), _mstate(NotInitialized), _fw_config(0),
   _verbose(false),_buffer(2048), _idle_pedestal(0),
   _idle_signal(0), _idle_send_data(0), _laser_delay(0),
   _cal_counter(0), _cal_delay(85), _cal_nevts(100), _cal_npts(0), _cal_ifirst(0), _cal_ilast(0),
   _ithmain(150), _ithdelta(15), _compctrl(0), _Vfp(0),
   _nevts(0)
{
    int i;
    for (i=0;i<256;i++)
    {
        _pedestals[i] = randgauss(0.,10.);
        _histogram[i] = 0;
        _ithcorr[i] = 0;
        _mask[i] = 0;
    }
    _idle_pedestal = g_idle_add((GSourceFunc) _generate_pedestals, this);

    vout_par[0] = time_offset;
    vout_par[1] = tau1(_Vfp);;
    vout_par[2] = 8;
    vout_par[3] = 1;

    double vmax = beetle_max(vout_par);
    vout_par[3] = 1./vmax;
}

void TheMachine::init(GIOChannel *fd)
{
    PDEBUG("Machine: Initializing the machine");
    state(Alibava::Reset);
    if (alibava_states.empty())
    {
        alibava_states[0x0] = "None";
        alibava_states[0x1] = "Reset";
        alibava_states[0x2] = "Waiting";
        alibava_states[0x3] = "BeetleConfig";
        alibava_states[0x4] = "Calibration";
        alibava_states[0x5] = "TriggerInConfig";
        alibava_states[0x6] = "LaserSync";
        alibava_states[0x7] = "LaserAcq";
        alibava_states[0x8] = "LaserReading";
        alibava_states[0x9] = "RadSourceAcq";
        alibava_states[0xa] = "RadSourceReading";
        alibava_states[0xb] = "PedestalAcq";
        alibava_states[0xc] = "PedestalReading";
        alibava_states[0xd] = "AlibavaConf";
        alibava_states[0xe] = "TriggerCounter";
        alibava_states[0xf] = "Connect";
        alibava_states[0x10] = "NewData";
        alibava_states[0x20] = "SlowControlError";
        alibava_states[0x30] = "ADCControlError1";
        alibava_states[0x40] = "ADCControlError2";
        alibava_states[0x50] = "ADCControlError";
        alibava_states[0x60] = "LaserAcqError";
        alibava_states[0x80] = "SDRAMError";
        alibava_states[0x81] = "Reset; fw version 2";
        alibava_states[0x81] = "Reset; fw version 3";
        alibava_states[0x8000] = "EndOfData";
    }
    //    send_state(fd);
}

void TheMachine::set_fw_version(int x)
{
    fw_version = x;
}


int TheMachine::get_nbits()
{
    int nb = 0;
    if ( !fw_version )
    {
        nb = 2096;
    }
    else
    {
        switch ( get_chip_mask() & 3 )
        {
            case 3:
                nb = 4656;
                break;
            default:
                nb = 2352;
                break;
        }
    }
    return nb;
}

void TheMachine::mstate(MachineState s)
{
    PDEBUG("Setting machine state: %s", TheMachine::sstate[s]);
    _mstate = s;
}

int TheMachine::_send_signal(TheMachine *m)
{
    if ( m->_nsent >= m->_nevts)
    {
        m->_idle_signal = 0;
        return 0;
    }
    else
    {
        m->send_data_block(m->_fd, BlockSize);
        m->_nsent++;
        return 1;
    }
}

int TheMachine::_generate_pedestals(TheMachine *m)
{
    if (m->_buffer.full() )
    {
    	PDEBUG("Buffer full. Stopping Pedestal generation\n");
        //info_msg("Stopping pedestal generation function");
        m->_idle_pedestal = 0;
        return 0;
    }
    //PDEBUG("Generating pedestal\n");
    m->create_data();
    return 1;
}

int TheMachine::get_chip_number()
{
    int ichip = 0;
    switch  ( get_chip_mask() & 3 )
    {
        case 1:
            ichip = 0;
            break;
        case 2:
            ichip = 1;
            break;
        default:
            ichip = (int)(2.*randuniform());
            break;
    }
    return ichip;
}



int TheMachine::_generate_signal(TheMachine *m)
{
    if (m->_memory.size()>=m->_nevts)
    {
        //info_msg("Stopping data generation function");
        m->_idle_signal = 0;
        m->set_to_wait((GIOChannel *)(m->tmp_data));
        return 0;
    }
    DataBlock *blk = m->get_data();
    EventDataBlock *evt = (EventDataBlock *)blk->data();
    double tim = randuniform();
    double twindow = (m->_fw_config & 2) ? 100.0 : 25.0;
    double xx = (twindow*tim);
    double tfact = beetle(&xx, m->vout_par);
    tim *= twindow/100.0;
    unsigned short ipart = (unsigned short)(tim * 65535);
    unsigned short fpart = (unsigned short)(tim*65535 - ipart);
    unsigned int   tval  = fpart + ipart;

    PDEBUG("Generating event signal\n");


//    {
//
//        unsigned short _fpart = tval & 0xffff;
//        short _ipart = ( tval & 0xffff0000)>>16;
//        if (_ipart<0)
//            _fpart *= -1;
//        double _tt = 100.0*(_ipart + (_fpart/65535.));
//        std::cout << "T0: " << tim*100. << " Tcode: " << _tt << std::endl
//        << "\tipart " << ipart << "/" << _ipart << " fpart: " << fpart << "/" << _fpart << std::endl;
//
//    }

    int ichip = m->get_chip_number();
    double polarity=1.;
    double T = randgauss(10., 2.);
    int chan = 127;
    double val = 0;
    switch (m->daq_type) {
        case source:
            chan = (int)(128.*randuniform());
            val = tfact*landau(30000.0, 500.);
            break;
        case laser:
            chan = 50;
            switch ( m->chip_mask )
            {
                case 2:
                case 3:
                    ichip = 1;
                    break;
                default:
                    ichip = 0;
                    break;
            }
            val = m->get_laser_amplitude(); //randgauss(20000., 500.);
            //std::cout << "machine: delay " << m->_laser_delay << " -> " << 0.006*polarity*val << std::endl;
            break;
        default:
            break;
    }
    val *= (polarity*0.006);
    if (val>400.)
        val = 400.;
    if (val<-400)
        val = -400.;

    // Fill the data
    if (clock_0 == 0)
        clock_0 = time(0);
    evt->clock = (unsigned int) ((clock() - clock_0 ));
    evt->time = tval;
    evt->temp = (unsigned short)((T+39.8)/0.12);
    evt->temp &=0x7FF;
    if ( m->daq_type!=pedestal )
    {
        if (m->daq_type != laser )
            evt->chip[ichip].signal[chan]  += (short)(val);
        else
        {
            evt->chip[ichip].signal[chan-2] += (short)(0.1*val);
            evt->chip[ichip].signal[chan+2] += (short)(0.1*val);
            evt->chip[ichip].signal[chan-1] += (short)(0.3*val);
            evt->chip[ichip].signal[chan+1] += (short)(0.3*val);
            evt->chip[ichip].signal[chan] += (short)(val);
        }
    }
    m->_memory.push_back(evt);
    return 1;
}

unsigned char TheMachine::read_command(GIOChannel *fd)
{
    GIOStatus status;
    GError *error=0;
    unsigned char code=0;
    _error = 0;
    gsize nread = 0;
    int rc=0;

    status = g_io_channel_read_chars(fd,(gchar *)&code, sizeof(unsigned char), &nread, &error);
    if ( error )
    {
        err_msg( _("Error sending the command: %s"), error->message);
        g_error_free(error);
        error = 0;
        _error = status;
        return code;
    }
    //rc = g_socket_receive(fd, (gchar *)&code, sizeof(unsigned char), 0, &error);
    if (nread!=sizeof(unsigned char))
    {
        err_msg( _("Error sending the command"));
        _error = rc<0 ? rc : EAGAIN;
    }
    return code;
}

unsigned short TheMachine::read_word(GIOChannel *fd)
{
    GIOStatus status;
    gsize nread;
    GError *error = 0;
    unsigned short code=0;
    _error = 0;
    int rc = 0;

    do
    {
        status = g_io_channel_read_chars(fd,(gchar *)&code, sizeof(unsigned short), &nread, &error);
        if (status == G_IO_STATUS_AGAIN || error )
            break;
        g_usleep(100);

    } while ( status == G_IO_STATUS_AGAIN && error == 0);
    if ( error )
    {
        err_msg( _("Error reading the word: %s"), error->message);
        g_error_free(error);
        error = 0;
        _error = status;
        return code;
    }
    //rc = g_socket_receive(fd, (gchar *)&code, sizeof(unsigned short), 0, &error);
    if (nread!=sizeof(unsigned short))
    {
        _error = status;
        err_msg(_("Apparently an incomplete read: status=%d"), status);
    }
    return code;
}

void TheMachine::reset()
{
    PDEBUG(_("Machine: reset"));
}

void TheMachine::hand_shake(GIOChannel *fd)
{
    mstate(WaitingForHandshake);
    PDEBUG(_("Machine::hand_shake"));
}

void TheMachine::set_to_wait(GIOChannel *fd)
{
    PDEBUG(_("Setting Waiting state"));
    _state = Alibava::Waiting;
    mstate(Waiting);
    send_state(fd, Alibava::Waiting);
}
void TheMachine::send_state(GIOChannel *fd, unsigned char s)
{
    GIOStatus status;
    GError *error=0;
    gsize nread;
    int rc;
    PDEBUG(_("Sending state %s"),alibava_states[s & 0xf].c_str());
    _error = 0;
    _state = 0xf & s;
    status = g_io_channel_write_chars(fd, (gchar *)&s, sizeof(unsigned char), &nread, &error);
    if ( error )
    {
        err_msg( _("Error sending the state: %s"), error->message);
        g_error_free(error);
        error = 0;
        _error = status;
        return;
    }
    //rc = g_socket_send(fd, (gchar *)&s, sizeof(unsigned char), 0, &error);
    if ( nread != sizeof(unsigned char))
        _error = status;

    //fsync(fd);
}

int TheMachine::send_data_block(GIOChannel *fd, int sz)
{
    GError *err = 0;
    gsize nsent = 0;
    PDEBUG(_("Sending data block of size %d"),sz);
    if (_memory.empty())
    {
        info_msg(_("memory empty... no more events to serve"));
        return -1;
    }

    int iblk, nblk = 1; //sz/BlockSize;
    PDEBUG(_("...number of blocks to send: %d"),nblk);
    for (iblk=0; iblk<nblk; iblk++)
    {
        EventDataBlock *blk = _memory.back();
        PDEBUG(_("...sending block %d"),iblk);
        g_io_channel_write_chars(fd, (gchar *)&blk->clock, sizeof(unsigned int), &nsent, &err);
        //g_socket_send (fd, (gchar *)&blk->time, sizeof(unsigned int), 0, &err);
        if ( err ) {
            err_msg(_("Error sending clock: %s"), err->message);
            g_error_free(err);
            err = 0;
        }
        g_io_channel_write_chars(fd, (gchar *)&blk->time, sizeof(unsigned int), &nsent, &err);
        //g_socket_send (fd, (gchar *)&blk->time, sizeof(unsigned int), 0, &err);
        if ( err ) {
            err_msg(_("Error sending time: %s"), err->message);
            g_error_free(err);
            err = 0;
        }
        g_io_channel_write_chars(fd, (gchar *)&blk->temp, sizeof(unsigned short), &nsent, &err);
        //g_socket_send (fd, (gchar *)&blk->temp, sizeof(unsigned short), 0, &err);
        if ( err ) {
            err_msg(_("Error sending temp.: %s"), err->message);
            g_error_free(err);
            err = 0;
        }
        if ( fw_version == 0)
        {
            g_io_channel_write_chars (fd, (gchar *)blk->chip[0].signal, 128*sizeof(unsigned short), 0, &err);
            if ( err ) {
                err_msg(_("Error sending chip 1 data block: %s\n"), err->message);
                g_error_free(err);
                err = 0;
            }
            g_io_channel_write_chars (fd, (gchar *)blk->chip[1].signal, 128*sizeof(unsigned short), 0, &err);
            if ( err ) {
                err_msg(_("Error sending chip 2 data block: %s\n"), err->message);
                g_error_free(err);
                err = 0;
            }
        }
        else
        {
            switch ( get_chip_mask() & 3)
            {
                case 1:
                    g_io_channel_write_chars (fd, (gchar *)blk->chip, sizeof(ChipBlock), 0, &err);
                    if ( err ) {
                        err_msg(_("Error Sending ChipBlock: %s\n"), err->message);
                        g_error_free(err);
                        err = 0;
                    }
                    break;
                case 2:
                    g_io_channel_write_chars (fd, (gchar *)(blk->chip+1), sizeof(ChipBlock), 0, &err);
                    if ( err ) {
                        err_msg(_("Error Sending ChipBlock: %s\n"), err->message);
                        g_error_free(err);
                        err = 0;
                    }
                    break;
                default:
                    g_io_channel_write_chars (fd, (gchar *)blk->chip, 2*sizeof(ChipBlock), 0, &err);
                    if ( err ) {
                        err_msg(_("Error Sending ChipBlock: %s\n"), err->message);
                        g_error_free(err);
                        err = 0;
                    }
                    break;
            }
        }
//        fsync(fd);
        _memory.pop_back();
        delete blk;
    }
    if (_memory.empty())
    {
        PDEBUG("Memory is empty. Going to wait state");
        set_to_wait(fd);
    }
    return 0;
}

/*
 * Here we process anything the user writes.
 * There are 2 parts:
 *   1) We check the internal state of the machine. If it is in
 *      anything other than idle, that means it is in the middle of
 *      processing a command. We try to resume where it left last
 *      time.
 *   2) It is in idle state. That means the user wants to change the
 *      state of the machine. We check for the new state mode and if
 *      valid we set the internal state to the corresponding one
 */
int TheMachine::process_state(GIOChannel *fd)
{
    int iok;
    unsigned char code;
    unsigned short usval;
    GError *__error = 0;

    bool check_command = false;
    _error = 0;
    switch ( mstate() )
    {
        case WaitingForHandshake:
            code = read_command(fd);
            VERBOSE("WaitingForHandshake: get code %02x\n", code);
            if (error())
                return -1;

            if (code!=Alibava::Reset)
            {
                err_msg(_("Wrong handshake code"));
                _error = EINVAL;
            }
            else
            {
                PDEBUG(_("Received handshake"));
            }
            mstate(Initialized);
            set_to_wait(fd);
            break;

        case BeetleConfiguration:
            VERBOSE("BeetleConfiguration\n");
            iok = beetle_config(fd);
            if (iok<0)
                return -1;

            break;

        case BeetleConfigReady:
            // Get the chip mask
            VERBOSE("BeetleConfigReady\n");
            if ( fw_version )
            {
                chip_mask = read_command(fd);
                PDEBUG(_("Getting chip mask: %d"), chip_mask);
            }
            mstate(BeetleConfiguration);
            send_state(fd, Alibava::NewData);
            break;

        case BeetleCalibration:
            VERBOSE("BeetleCalibration\n");
            iok = beetle_calib(fd);
            if (iok<0)
                return 0;

            break;

        case AlibavaConfReady:
            VERBOSE("AlibavaConfReady\n");
            _fw_config = read_word(fd);
            mstate(AlibavaConfEnd);
            send_state(fd, Alibava::NewData);
            break;

        case AlibavaConfEnd:
            VERBOSE("AlibavaConfEnd\n");
            usval = read_word(fd);
            set_to_wait(fd);
            break;

        case TriggerCountReady:
            VERBOSE("TriggerCountReady\n");
            usval = read_command(fd);
            if (!usval)
                ntrigger = 0;

            g_io_channel_write_chars (fd, (gchar *)&ntrigger, sizeof(unsigned short), 0, &__error);
            set_to_wait(fd);
            break;

        case TriggerCountEnd:
            VERBOSE("TriggerCountEnd\n");
            //        g_io_channel_write_chars (fd, (gchar *)&ntrigger, sizeof(unsigned short), 0, &__error);
            //        set_to_wait(fd);
            break;

        case TriggerIn:
            code = read_command(fd);
            PDEBUG("Trigger config is: %x", code);
            VERBOSE("TriggerIn: code=%02x\n", code);

            mstate(TriggerDACs);
            send_state(fd, Alibava::NewData);
            break;

        case TriggerDACs:
            VERBOSE("TriggerDACs\n");
            iok = trigger_config(fd);
            if (iok<0)
                return -1;

            break;

        case LaserSynchronization:
            VERBOSE("LaserSynchronization\n");
            iok = laser_synchronization(fd);
            if (iok<0)
                return -1;

            break;

        case PrepareAcquisition:
            _nevts = read_word(fd);
            VERBOSE("PrepareAcquisition: nevts=%d\n", _nevts);
            if (_nevts>0)
            {
                _memory.clear();
                tmp_data = (void *)fd;
                PDEBUG("Starting acquisition");
                _idle_signal = g_idle_add((GSourceFunc) _generate_signal, this);
                mstate(Acquiring);
            }
            else
            {
                send_state(fd, Alibava::SDRAMError);
                set_to_wait(fd);
            }
            break;

        case ChargeScanGetDelay:
            mstate(ChargeScanGetNevts);
            _cal_delay = read_word(fd);
            VERBOSE("ChargeScanGetDelay: delay=%d\n",_cal_delay);
             send_state(fd, Alibava::NewData);
            break;

        case ChargeScanGetNevts:
            mstate(ChargeScanGetNpts);
            _cal_nevts = read_word(fd);
            VERBOSE("ChargeScanGetNevts: nevts=%d\n", _cal_nevts);
            send_state(fd, Alibava::NewData);
            break;

        case ChargeScanGetNpts:
            mstate(ChargeScanGetIfirst);
            _cal_npts = read_word(fd);
            VERBOSE("ChargeScanGetNpts: Npts=%d\n", _cal_npts);
            send_state(fd, Alibava::NewData);

            break;

        case ChargeScanGetIfirst:
            mstate(ChargeScanGetIlast);
            _cal_ifirst = read_word(fd);
            VERBOSE("ChargeScanGetIfirst: first=%d\n", _cal_ifirst);
            send_state(fd, Alibava::NewData);
            break;

        case ChargeScanGetIlast:
            mstate(ChargeScanGetIlast);
            _cal_ilast = read_word(fd);
            VERBOSE("ChargeScanGetIlast: last=%d\n", _cal_ilast);
            handle_charge_scan(fd);
            break;


        case WaitReadAck:
            code = read_command(fd);
            VERBOSE("WaitReadAck: code=%02x\n", code);
            mstate(ReadingData);
            _nsent=0;
            _fd = fd;
            _idle_signal = g_idle_add((GSourceFunc)_send_signal, (gpointer)this);
            break;

        case Acquiring:
//            VERBOSE("Acquiring\n");
            break;

        case SoftwareReset:
            VERBOSE("SoftwareReset\n");
            check_command = true;
            break;

        default:
            check_command = true;
            break;
    }

    if (!check_command)
        return 0;

    /**
     * Process the command code sent from the PC
     */
    code = read_command(fd);
    if (error() )
        return -1;

    unsigned char sfw;
    switch (code)
    {
        case Alibava::Connect:
            PDEBUG("Connect");
            VERBOSE("Command: Connect\n");
            if (need_init() )
            {
                init(fd);
                hand_shake(fd);
            }
            sfw = (fw_version & 0xf);
            sfw = reverse_bit(sfw);
            send_state(fd, sfw | state());

            break;
        case Alibava::Reset:
            VERBOSE("Command: Reset\n");
            PDEBUG("Reset");
            break;
        case Alibava::AlibavaConf:
            VERBOSE("Command: AlibavaConf\n");
            send_state(fd, Alibava::AlibavaConf);
            mstate(AlibavaConfReady);
            break;
        case Alibava::TriggerCounter:
            VERBOSE("Command: TriggerCounter\n");
            send_state(fd, Alibava::TriggerCounter);
            mstate(TriggerCountReady);
            break;
        case Alibava::Waiting:
            VERBOSE("Command: Waiting\n");
            PDEBUG("Waiting");
            break;
        case Alibava::BeetleConfig:
            VERBOSE("Command: BeetleConfig (send NewData)\n");
            PDEBUG("BeetleConfig");
            indx = 0;
            if (! fw_version )
            {
                mstate(BeetleConfiguration);
                indx = 0;
                send_state(fd, Alibava::BeetleConfig);
                send_state(fd, Alibava::NewData);
            }
            else
            {
                send_state(fd, Alibava::BeetleConfig);
                mstate(BeetleConfigReady);
            }
            break;
        case Alibava::Calibration:
            VERBOSE("Command: Calibration\n");
            PDEBUG("Calibration");
            mstate(BeetleCalibration);
            send_state(fd, Alibava::Calibration);
            send_state(fd, Alibava::NewData);
            break;

        case Alibava::TriggerInConfig:
            VERBOSE("Command: TriggerInConfig\n");
            PDEBUG("TriggerConfig");
            mstate(TriggerIn);
            send_state(fd, Alibava::TriggerInConfig);
            break;
        case Alibava::LaserSync:
            VERBOSE("Command: LaserSync\n");
            PDEBUG("LaserSync");
            mstate(LaserSynchronization);
            indx = 0;
            send_state(fd, Alibava::LaserSync);
            send_state(fd, Alibava::NewData);
            break;
        case Alibava::LaserAcq:
            VERBOSE("Command: LaserAcq\n");
            PDEBUG("LaserAcq");
            mstate(PrepareAcquisition);
            daq_type = laser;
            send_state(fd, Alibava::LaserAcq);

            break;
        case Alibava::RadSourceAcq:
            VERBOSE("Command: RadSourceAcq\n");
            PDEBUG("RadSourceAcq");
            mstate(PrepareAcquisition);
            daq_type = source;
            send_state(fd, Alibava::RadSourceAcq);
            break;

        case Alibava::PedestalAcq:
            VERBOSE("Command: PedestalAcq\n");
            PDEBUG("PedestalAcq");
            mstate(PrepareAcquisition);
            daq_type = pedestal;
            send_state(fd, Alibava::PedestalAcq);
            break;
        case Alibava::LaserReading:
        case Alibava::RadSourceReading:
        case Alibava::PedestalReading:
            VERBOSE("Command: Reading\n");
            PDEBUG("Reading");
            usval = get_nbits();
            send_state(fd, code);
            g_io_channel_write_chars (fd, (gchar *)&_nevts, sizeof(unsigned short), 0, &__error);
            g_io_channel_write_chars (fd, (gchar *)&usval, sizeof(unsigned short), 0, &__error);
            //fsync(fd);
            mstate(WaitReadAck);
            break;

        case Alibava::ChargeScan:
            VERBOSE("Command: ChargeScan\n");
            PDEBUG("ChargeScan");
            mstate(ChargeScanGetDelay);
            send_state(fd, code);
            send_state(fd, Alibava::NewData);
            break;

        case Alibava::NewData:
            VERBOSE("Command: NewData\n");
            PDEBUG("NewData");
            break;
        case Alibava::SlowControlError:
            VERBOSE("Command: SlowControlError\n");
            PDEBUG("SlowControlError");
            break;
        case Alibava::SftReset:
            VERBOSE("Command: SoftwareReset\n");
            PDEBUG("SoftwareReset");
            nreset--;
            if (!nreset)
            {
                mstate(NotInitialized);
                nreset=4;
            }
            else
                mstate(SoftwareReset);
            break;
        default:
            VERBOSE("Command: Unknown\n");
            PDEBUG("Unknown code");
            break;

    }

    return 0;
}


int TheMachine::beetle_config(GIOChannel *fd)
{
    double vout;
    unsigned short word;
    PDEBUG("reading beetle config word %d",indx);
    _error = 0;
    word = read_word(fd);
    if (error() )
        return -1;

    PDEBUG("...word is %x",word);
    VERBOSE("Beetle word: %4x\n", word);
    if (word == Alibava::EndOfData)
    {
        PDEBUG("Got end of data");
        send_state(fd, Alibava::Waiting);
        mstate(Initialized);
    }
    else
    {
        // decode the word and store it
        indx = (word & 0xff00)>>8;
        if (indx<0|| indx>22)
        {
            PDEBUG("Wrong index");
            err_msg("Wrong Beetle conf index: %x", word);
            send_state(fd, Alibava::SlowControlError);
            state(Alibava::Waiting);
            _error = EINVAL;
        }
        else
        {
            _beetle_conf[indx] = word;
            if ( indx < 20 )
            {
                for (int i=0;i<3;i++)
                    _ncnt[i] = 0;

                switch (indx)
                {
                    case 8:
                        _ithmain = word & 0xff;
                        //printf("Ithmain: %d\n", _ithmain);
                        break;
                    case 7:
                        _ithdelta = word & 0xff;
                        break;
                    case 4:
                        _Vfp = word & 0xff;

                        vout_par[1] = tau1(_Vfp);
                        vout_par[0] = 0;
                        vout_par[3] = 1;
                        vout = beetle_max(vout_par);
                        vout_par[0] = time_offset;
                        vout_par[3] = 1.0/vout;
                        break;
                    case 19:
                        _compctrl = word & 0xff;
                        break;
                    default:
                        break;
                }
            }
            else
            {
                int j = indx - 20;
                switch(indx)
                {
                    case 20:
                        _ithcorr[_ncnt[j]] = word & 0xff;
                        break;
                    case 21:
                        word &= 0xff;
                        //printf("mask %d: ", _ncnt[j]);
                        for (int ibit=0; ibit<8; ibit++)
                        {
                            int offs = _ncnt[j]*8 + ibit;
                            _mask[offs] = (word & (1<<ibit)) ? 1 : 0;
                            //printf("%d", _mask[offs] );
                        }
                        //printf("\n");
                        break;
                    default:
                        break;
                }
                _ncnt[j]++;
                if (indx==21)
                    _ncnt[j] = _ncnt[j] %16;
                else if (indx==20)
                    _ncnt[j] = _ncnt[j] %128;
            }
            PDEBUG("Sending NewData word");
            send_state(fd, Alibava::NewData);
        }
    }
    return 0;
}


/**
 * From the data, ithmain slope is 0.00135 dac/electron, the offset
 * is such that, ideally, the read band in in 128 dac units:
 *
 * Thrs (DAC) = 128 + 0.00135 * Q(e-)
 *
 * Then the threshold correction has a slope of 8 dac units, that is
 * one unit of corr. moves the threshold by 8 units.
 *
 */
int TheMachine::handle_charge_scan(GIOChannel *fd)
{
    int i;
    GError *err = 0;
    GIOStatus status;
    gsize nsent = 0;
    unsigned int step;

    if (_cal_ifirst > _cal_ilast)
    {
        unsigned short tmp = _cal_ilast;
        _cal_ilast = _cal_ifirst;
        _cal_ifirst = tmp;
    }
    if (_cal_npts>256)
        _cal_npts = 256;

    if (_cal_npts>1)
    {
        step = (_cal_ilast-_cal_ifirst)/(_cal_npts - 1);
        if (step==0)
            step = 1;
    }
    else
        step = 1;


    int ipol = (_compctrl & 2) ? 1 : 0;

    // Reset histogram
    for (i=0; i<_cal_npts; i++)
        _histogram[i] = 0;

    /*
     * Generate the calibration events
     * We should look at the mask, ithmain, etc from the
     * beetle configuration.
     */
    int nchip = (chip_mask==3 ? 2 : 1);
    int wchip = (chip_mask==3 ? 0 : chip_mask-1);
    int woffs = wchip * 128;
    int nchan = nchip * 128;
    int ival  = _cal_ifirst;
    std::vector<int> chan_list;
    std::vector<int>::iterator itchan;
    for (i=0;i<nchan;i++)
    {
        int jchan = i + woffs;
        if (_mask[jchan])
            continue;

        //printf("Accept channel %d\n", jchan);
        chan_list.push_back(jchan);
    }

    for (int ipoint=0; ipoint<_cal_npts; ipoint++)
    {
        ival = _cal_ifirst + ipoint*step;
        double Q = 1024.0*ival;
        for (int ievt = 0; ievt < _cal_nevts; ievt++)
        {
            short sign = -2*( (_cal_counter++) % 2)+1;
            double tfact = 1.0/150.0; //generate_calib(_cal_delay)/350.0; // 350 electrons/thrs dac
            double Qr = randgauss(Q, 3500.0); // Add the noise
            double Qd = tfact*Qr;
            int ntrg = 0;
            for (itchan = chan_list.begin(); itchan!=chan_list.end(); ++itchan)
            {
                int jchan = *itchan;
                int ichan = jchan - woffs;

                double S    = (-2*(ichan%2) + 1)*sign;
                double Qeff = S*Qd;
//                if (fabs(Qr/3500)<1)
//                {
//                    ntrg++;
//                    continue;
//                }
                if ( Qeff<0 )
                    Qeff = 0.25*fabs(Qeff);

                //Qeff = 256 - Qeff;

                double Th = _ithmain - (comparator_offset[jchan]-150);
                double Th0 = Th;
                double corr = 3*_ithcorr[jchan]+40;
                Th -= corr;
                if ( Th <  0)
                {
                    if ( Qeff < 0 && Qeff < Th )
                        ntrg++;
                }
                else
                {
                    if (Qeff > Th)
                        ntrg++;
                }
                //printf("ichan %d Ith %d Trim %d Th0 %f Th_corr %f Q %f Qeff %f ntrg %d\n", jchan, _ithmain, _ithcorr[jchan], Th0, Th, Q, Qeff, ntrg);
//                std::cout << "Qeff " << Qeff
//                          << " Th " << Th
//                          << " corr " << corr
//                          << " ithcorr[" << jchan << "]=" << _ithcorr[jchan]
//                          << " ntrg " << ntrg << " - " << _histogram[ipoint] << std::endl;
            }
            if (ntrg)
                _histogram[ipoint]++;
        }
    }

    /*
     * Return the number of points and the counters
     */
    status = g_io_channel_write_chars (fd, (gchar *)&_cal_npts, sizeof(unsigned short), &nsent, &err);
    if (err)
    {
        _error = err->code;
        err_msg(_("Error sending number of points: %s\n"), err->message);
        g_error_free(err);
        err = 0;
    }
    status = g_io_channel_write_chars (fd, (gchar *)_histogram, _cal_npts*sizeof(unsigned short), &nsent, &err);
    if (err)
    {
        _error = err->code;
        err_msg(_("Error sending counters: %s\n"), err->message);
        g_error_free(err);
        err = 0;
    }
    /*
     * Back to waiting state
     */
    send_state(fd, Alibava::Waiting);
    mstate(Initialized);
    return 0;
}

double TheMachine::generate_calib(double t)
{
    double xx = (t-68.)/20.;
    double tfact = xx < 0.0 ? 0.0 : xx*exp(1.0-xx);
    return tfact;
}
int TheMachine::beetle_calib(GIOChannel *fd)
{
    GError *__error = 0;
    int rc=0;
    gsize nsent = 0;
    GIOStatus status;
    unsigned short delay=88, word;

    if ( fw_version )
    {
        PDEBUG("reading beetle calibration delay");
    }
    else
    {
        PDEBUG("reading beetle calibration pulse (fw=0)");
    }

    _error = 0;
    word = read_word(fd);
    if (error() )
        return -1;

    PDEBUG("...value is %x", word);
    if (word == Alibava::EndOfData)
    {
        PDEBUG("Got end of data");
        send_state(fd, Alibava::Waiting);
        mstate(Initialized);
    }
    else
    {
        double tfact = 1.;
        if ( fw_version )
        {
            // The previous word was the delay. Let's get now
            // the pulse value
            // TODO: shall we simulate also the delay ?
            PDEBUG("reading calibration pulse (fw=1)");
            delay = word;
            word = read_word(fd);
            if ( error() )
            {
                err_msg(_("Error reading pulse value"));
                return -1;
            }
            tfact = generate_calib(delay);
            PDEBUG(_("...value is %x"), word);
        }
        // We have to generate the data
        DataBlock *data = get_data();
        EventDataBlock *evt = (EventDataBlock *)data->data();

        double charge = (1024.*word)*0.0025; // charge in mV
        int value = int(tfact*charge/2.);
        double T = randgauss(10., 2.);
        evt->clock = 0;
        evt->time = 0;
        evt->temp = (0x7ff & (unsigned short)((T+39.8)/0.12));
        int ic;
        int ichip;
        short sign = -2*( (_cal_counter++) % 2)+1;
        for (ichip=0; ichip<2; ichip++)
        {
            if ( !(get_chip_mask() & 1<<ichip) )
                continue;

            for (ic=0; ic<128; ic+=2)
            {
                evt->chip[ichip].signal[ic] += sign*value;
                evt->chip[ichip].signal[ic+1] += -sign*value;
            }
        }
        if ( fw_version == 0)
        {
            for (int i=0; i<2; i++)
            {
                status = g_io_channel_write_chars (fd, (gchar *)evt->chip[i].signal, 128*sizeof(unsigned short), &nsent, &__error);
                if ( __error ) {
                    _error = __error->code;
                    err_msg(_("Error sending time: %s\n"), __error->message);
                    g_error_free(__error);
                    __error = 0;
                }
            }
        }
        else
        {
            int nchip = (chip_mask==3 ? 2 : 1);
            int wchip = (chip_mask==3 ? 0 : chip_mask-1);
            status = g_io_channel_write_chars (fd, (gchar *)&evt->chip[wchip], nchip*sizeof(ChipBlock), &nsent, &__error);
            if ( __error ) {
                _error = __error->code;
                err_msg(_("Error sending time: %s\n"), __error->message);
                g_error_free(__error);
                __error = 0;
            }
        }
        //fsync(fd);
        send_state(fd, Alibava::NewData);
        delete data;
    }
    return rc;
}

double TheMachine::get_laser_amplitude()
{
    double T = _laser_delay - 150.;
    double xx = T/50.;
    double value = T < 0 ? 0 : 20000.*xx*exp(1-xx);
    return value;
}

int TheMachine::laser_synchronization(GIOChannel *fd)
{
    gsize rc=0;
    unsigned short delay;
    GError *__error = 0;

    PDEBUG(_("reading laser delay register"));
    _error = 0;

    g_io_channel_read_chars(fd, (gchar *)&delay, sizeof(unsigned short), &rc, &__error);
    if (rc!=sizeof(unsigned short))
    {
        _error = rc<0 ? rc : EAGAIN;
    }
    if (error() )
        return -1;

    PDEBUG(_("delay register is %x"), delay);

    if (delay == 0x8000)
    {
        PDEBUG(_("Got end of data"));
        send_state(fd, Alibava::Waiting);
        mstate(Initialized);
    }
    else
    {
        // We have to generate the data
        _laser_delay = delay;
        double value = get_laser_amplitude();
        DataBlock *data = get_data();
        EventDataBlock *evt = (EventDataBlock *)data->data();
        double T = randgauss(10., 2.);
        value *= 0.006;
        evt->clock = 0;
        evt->time = 0;
        evt->temp = (0x7ff & (unsigned short)((T+39.8)/0.12));
        int ichip = get_chip_number();
        evt->chip[ichip].signal[100] += (unsigned short)value;

        if ( fw_version == 0)
        {
            rc = g_io_channel_write_chars (fd, (gchar *)evt->chip[0].signal, 128*sizeof(unsigned short), 0, &__error);
            rc = g_io_channel_write_chars (fd, (gchar *)evt->chip[1].signal, 128*sizeof(unsigned short), 0, &__error);
        }
        else
        {
            switch ( get_chip_mask() & 3 )
            {
                case 1:
                    rc = g_io_channel_write_chars (fd, (gchar *)evt->chip, sizeof(ChipBlock), 0, &__error);
                    break;
                case 2:
                    rc = g_io_channel_write_chars (fd, (gchar *)(evt->chip+1), sizeof(ChipBlock), 0, &__error);
                    break;
                default:
                    rc = g_io_channel_write_chars (fd, (gchar *)evt->chip, 2*sizeof(ChipBlock), 0, &__error);
                    break;
            }
        }
        //fsync(fd);
        send_state(fd, Alibava::NewData);
        delete data;
    }
    return rc;
}

int TheMachine::trigger_config(GIOChannel *fd)
{
    int rc=0;
    unsigned short word;

    // we have to read the DACs
    _error = 0;
    word = read_word(fd);
    if (error() )
        return -1;

    if (word == Alibava::EndOfData)
    {
        PDEBUG(_("Got end of data"));
        send_state(fd, Alibava::Waiting);
        mstate(Initialized);
    }
    else
    {
        // We have to generate the data
        PDEBUG(_("DAC thrshold: %x"),word);
        send_state(fd, Alibava::NewData);
    }
    rc = error();
    return rc;
}

DataBlock *TheMachine::get_data()
{
    if (_buffer.empty() )
    {
        create_data();
        if (_idle_pedestal==0)
        {
            //info_msg("Starting pedestal generation function");
            _idle_pedestal = g_idle_add((GSourceFunc) _generate_pedestals, this);
        }
    }
    return _buffer.get_block();
}

/**
 * Creates pedestals
 *
 */
void TheMachine::create_data()
{
    if (_buffer.full() )
        return;

    DataBlock *data = new DataBlock( sizeof(EventDataBlock) );
    EventDataBlock *evt = (EventDataBlock *)data->data();
    int i,iw;
    double T = randgauss(10., 2.);
    evt->clock = 0;
    evt->time = 0;
    evt->temp = (0x7ff & (unsigned short)((T+39.8)/0.12));

    /**
    for (i=0;i<8;i++)
    {
        evt->chip[0].header[i] = 1024*i/16;
        evt->chip[1].header[7+i] = 1024*i/16;
    }
    for (;i<16;i++)
    {
        evt->chip[0].header[i] = (1024*randuniform());
        evt->chip[1].header[i-8] = (1024*randuniform());
    }
    **/
    evt->chip[0].header[0] = 630.0;
    evt->chip[1].header[0] = 630.0;
    evt->chip[0].header[1] = 512.0;
    evt->chip[1].header[1] = 512.0;
    evt->chip[0].header[2] = 630.0;
    evt->chip[1].header[2] = 630.0;
    for (i=3; i<16; i++)
    {
        evt->chip[0].header[i] = 512.0;
        evt->chip[1].header[i] = 512.0;
    }

    double ped, cmmd;
    cmmd = randgauss(0.,5.);
    for (iw=0, i=0; i<128;i++, iw++)
    {
        ped = randgauss(_pedestals[i], 5.) + cmmd;
        evt->chip[0].signal[iw] = (unsigned int) (ped/2.0+512.0);
    }

    cmmd = randgauss(0.,3.);
    for (i=0; i<128;i++, iw++)
    {
        ped = randgauss(_pedestals[128+i], 7.) + cmmd;
        evt->chip[1].signal[i] = (unsigned int) (ped/2.0+512.0);
    }
    _buffer.push_block(data);
}

/**
 Random numbers
 **/
#include <cmath>
#include <cstdlib>

static double randuniform()
{
    double x;
    x = ((double) rand()) / ((double) RAND_MAX);
    return x;
}

static double randgauss(double mean, double stdev)
{
    static int haveCachedNormal = 0;
    static double cachedNormal = 0.;
    static double pMean = 0., pStdDev = 0.;

    if (mean != pMean || pStdDev != stdev)
    {
        haveCachedNormal = 0;
        pMean = mean;
        pStdDev = stdev;
    }

    if (haveCachedNormal == 1)
    {
        haveCachedNormal = 0;
        return (cachedNormal * pStdDev + pMean);
    }
    else
    {
        for (;;)
        {
            double u1 = randuniform();
            double u2 = randuniform();
            double v1 = 2 * u1 - 1;
            double v2 = 2 * u2 - 1;
            double w = (v1 * v1) + (v2 * v2);

            /*
             We actually generate two IID normal distribution variables.
             We cache the one & return the other. */
            if (w <= 1)
            {
                double y = sqrt((-2 * log(w)) / w);
                double x1 = v1 * y;
                double x2 = v2 * y;

                haveCachedNormal = 1;
                cachedNormal = x2;
                return (x1 * pStdDev + pMean);
            }
        }
    }
    return mean;
}

static float __f[977] = { -2.244733, -2.204365, -2.168163, -2.135219, -2.104898, -2.07674, -2.050397, -2.025605, -2.00215,
        -1.979866, -1.958612, -1.938275, -1.91876, -1.899984, -1.881879, -1.864385, -1.847451, -1.83103, -1.815083, -1.799574,
        -1.784473, -1.769751, -1.755383, -1.741346, -1.72762, -1.714187, -1.701029, -1.68813, -1.675477, -1.663057, -1.650858,
        -1.638868, -1.627078, -1.615477, -1.604058, -1.592811, -1.581729, -1.570806, -1.560034, -1.549407, -1.538919, -1.528565,
        -1.518339, -1.508237, -1.498254, -1.488386, -1.478628, -1.468976, -1.459428, -1.449979, -1.440626, -1.431365, -1.422195,
        -1.413111, -1.404112, -1.395194, -1.386356, -1.377594, -1.368906, -1.360291, -1.351746, -1.343269, -1.334859, -1.326512,
        -1.318229, -1.310006, -1.301843, -1.293737, -1.285688, -1.277693, -1.269752, -1.261863, -1.254024, -1.246235, -1.238494,
        -1.230800, -1.223153, -1.215550, -1.207990, -1.200474, -1.192999, -1.185566, -1.178172, -1.170817, -1.163500, -1.156220,
        -1.148977, -1.141770, -1.134598, -1.127459, -1.120354, -1.113282, -1.106242, -1.099233, -1.092255, -1.085306, -1.078388,
        -1.071498, -1.064636, -1.057802, -1.050996, -1.044215, -1.037461, -1.030733, -1.024029, -1.01735, -1.010695, -1.004064,
        -0.997456, -0.990871, -0.984308, -0.977767, -0.971247, -0.964749, -0.958271, -0.951813, -0.945375, -0.938957, -0.932558,
        -0.926178, -0.919816, -0.913472, -0.907146, -0.900838, -0.894547, -0.888272, -0.882014, -0.875773, -0.869547, -0.863337,
        -0.857142, -0.850963, -0.844798, -0.838648, -0.832512, -0.82639, -0.820282, -0.814187, -0.808106, -0.802038, -0.795982,
        -0.78994, -0.783909, -0.777891, -0.771884, -0.765889, -0.759906, -0.753934, -0.747973, -0.742023, -0.736084, -0.730155,
        -0.724237, -0.718328, -0.712429, -0.706541, -0.700661, -0.694791, -0.688931, -0.683079, -0.677236, -0.671402, -0.665576,
        -0.659759, -0.65395, -0.648149, -0.642356, -0.63657, -0.630793, -0.625022, -0.619259, -0.613503, -0.607754, -0.602012,
        -0.596276, -0.590548, -0.584825, -0.579109, -0.573399, -0.567695, -0.561997, -0.556305, -0.550618, -0.544937, -0.539262,
        -0.533592, -0.527926, -0.522266, -0.516611, -0.510961, -0.505315, -0.499674, -0.494037, -0.488405, -0.482777, -0.477153,
        -0.471533, -0.465917, -0.460305, -0.454697, -0.449092, -0.443491, -0.437893, -0.432299, -0.426707, -0.421119, -0.415534,
        -0.409951, -0.404372, -0.398795, -0.393221, -0.387649, -0.38208, -0.376513, -0.370949, -0.365387, -0.359826, -0.354268,
        -0.348712, -0.343157, -0.337604, -0.332053, -0.326503, -0.320955, -0.315408, -0.309863, -0.304318, -0.298775, -0.293233,
        -0.287692, -0.282152, -0.276613, -0.271074, -0.265536, -0.259999, -0.254462, -0.248926, -0.243389, -0.237854, -0.232318,
        -0.226783, -0.221247, -0.215712, -0.210176, -0.204641, -0.199105, -0.193568, -0.188032, -0.182495, -0.176957, -0.171419,
        -0.16588, -0.160341, -0.154800, -0.149259, -0.143717, -0.138173, -0.132629, -0.127083, -0.121537, -0.115989, -0.110439,
        -0.104889, -0.099336, -0.093782, -0.088227, -0.08267, -0.077111, -0.07155, -0.065987, -0.060423, -0.054856, -0.049288,
        -0.043717, -0.038144, -0.032569, -0.026991, -0.021411, -0.015828, -0.010243, -0.004656, 9.34e-4, .006527, .012123, .017722,
        .023323, .028928, .034535, .040146, .045759, .051376, .056997, .06262, .068247, .073877, .079511, .085149, .09079, .096435,
        .102083, .107736, .113392, .119052, .124716, .130385, .136057, .141734, .147414, .1531, .158789, .164483, .170181, .175884,
        .181592, .187304, .193021, .198743, .204469, .210201, .215937, .221678, .227425, .233177, .238933, .244696, .250463,
        .256236, .262014, .267798, .273587, .279382, .285183, .290989, .296801, .302619, .308443, .314273, .320109, .325951,
        .331799, .337654, .343515, .349382, .355255, .361135, .367022, .372915, .378815, .384721, .390634, .396554, .402481,
        .408415, .414356, .420304, .42626, .432222, .438192, .444169, .450153, .456145, .462144, .468151, .474166, .480188,
                          .486218, .492256, .498302, .504356, .510418, .516488, .522566, .528653, .534747, .54085, .546962,
                          .553082, .55921, .565347, .571493, .577648, .583811, .589983, .596164, .602355, .608554, .614762, .62098,
                                  .627207, .633444, .639689, .645945, .65221, .658484, .664768, .671062, .677366, .68368, .690004,
                                  .696338, .702682, .709036, .7154, .721775, .72816, .734556, .740963, .747379, .753807, .760246,
                                  .766695, .773155, .779627, .786109, .792603, .799107, .805624, .812151, .81869, .825241, .831803,
                                  .838377, .844962, .85156, .85817, .864791, .871425, .878071, .884729, .891399, .898082, .904778,
                                  .911486, .918206, .92494, .931686, .938446, .945218, .952003, .958802, .965614, .972439, .979278,
                                  .98613, .992996, .999875, 1.006769, 1.013676, 1.020597, 1.027533, 1.034482, 1.041446, 1.048424,
                                  1.055417, 1.062424, 1.069446, 1.076482, 1.083534, 1.0906, 1.097681, 1.104778, 1.111889, 1.119016,
                                  1.126159, 1.133316, 1.14049, 1.147679, 1.154884, 1.162105, 1.169342, 1.176595, 1.183864,
                          1.191149, 1.198451, 1.20577, 1.213105, 1.220457, 1.227826, 1.235211, 1.242614, 1.250034, 1.257471,
                                  1.264926, 1.272398, 1.279888, 1.287395, 1.294921, 1.302464, 1.310026, 1.317605, 1.325203,
                                  1.332819, 1.340454, 1.348108, 1.35578, 1.363472, 1.371182, 1.378912, 1.38666, 1.394429, 1.402216,
                                  1.410024, 1.417851, 1.425698, 1.433565, 1.441453, 1.44936, 1.457288, 1.465237, 1.473206,
                          1.481196, 1.489208, 1.49724, 1.505293, 1.513368, 1.521465, 1.529583, 1.537723, 1.545885, 1.554068,
                                  1.562275, 1.570503, 1.578754, 1.587028, 1.595325, 1.603644, 1.611987, 1.620353, 1.628743,
                                  1.637156, 1.645593, 1.654053, 1.662538, 1.671047, 1.679581, 1.688139, 1.696721, 1.705329,
                                  1.713961, 1.722619, 1.731303, 1.740011, 1.748746, 1.757506, 1.766293, 1.775106, 1.783945,
                          1.79281, 1.801703, 1.810623, 1.819569, 1.828543, 1.837545, 1.846574, 1.855631, 1.864717, 1.87383,
                                  1.882972, 1.892143, 1.901343, 1.910572, 1.91983, 1.929117, 1.938434, 1.947781, 1.957158,
                          1.966566, 1.976004, 1.985473, 1.994972, 2.004503, 2.014065, 2.023659, 2.033285, 2.042943, 2.052633,
                                  2.062355, 2.07211, 2.081899, 2.09172, 2.101575, 2.111464, 2.121386, 2.131343, 2.141334, 2.15136,
                                  2.161421, 2.171517, 2.181648, 2.191815, 2.202018, 2.212257, 2.222533, 2.232845, 2.243195,
                                  2.253582, 2.264006, 2.274468, 2.284968, 2.295507, 2.306084, 2.316701, 2.327356, 2.338051,
                                  2.348786, 2.359562, 2.370377, 2.381234, 2.392131, 2.40307, 2.414051, 2.425073, 2.436138,
                          2.447246, 2.458397, 2.469591, 2.480828, 2.49211, 2.503436, 2.514807, 2.526222, 2.537684, 2.54919,
                                  2.560743, 2.572343, 2.583989, 2.595682, 2.607423, 2.619212, 2.63105, 2.642936, 2.654871,
                          2.666855, 2.67889, 2.690975, 2.70311, 2.715297, 2.727535, 2.739825, 2.752168, 2.764563, 2.777012,
                                  2.789514, 2.80207, 2.814681, 2.827347, 2.840069, 2.852846, 2.86568, 2.87857, 2.891518, 2.904524,
                                  2.917588, 2.930712, 2.943894, 2.957136, 2.970439, 2.983802, 2.997227, 3.010714, 3.024263,
                                  3.037875, 3.051551, 3.06529, 3.079095, 3.092965, 3.1069, 3.120902, 3.134971, 3.149107, 3.163312,
                                  3.177585, 3.191928, 3.20634, 3.220824, 3.235378, 3.250005, 3.264704, 3.279477, 3.294323,
                          3.309244, 3.32424, 3.339312, 3.354461, 3.369687, 3.384992, 3.400375, 3.415838, 3.431381, 3.447005,
                                  3.462711, 3.4785, 3.494372, 3.510328, 3.52637, 3.542497, 3.558711, 3.575012, 3.591402, 3.607881,
                                  3.62445, 3.641111, 3.657863, 3.674708, 3.691646, 3.70868, 3.725809, 3.743034, 3.760357, 3.777779,
                                  3.7953, 3.812921, 3.830645, 3.84847, 3.8664, 3.884434, 3.902574, 3.920821, 3.939176, 3.95764,
                                  3.976215, 3.994901, 4.013699, 4.032612, 4.051639, 4.070783, 4.090045, 4.109425, 4.128925,
                                  4.148547, 4.168292, 4.18816, 4.208154, 4.228275, 4.248524, 4.268903, 4.289413, 4.310056,
                          4.330832, 4.351745, 4.372794, 4.393982, 4.41531, 4.436781, 4.458395, 4.480154, 4.50206, 4.524114,
                                  4.546319, 4.568676, 4.591187, 4.613854, 4.636678, 4.659662, 4.682807, 4.706116, 4.72959,
                          4.753231, 4.777041, 4.801024, 4.825179, 4.849511, 4.87402, 4.89871, 4.923582, 4.948639, 4.973883,
                                  4.999316, 5.024942, 5.050761, 5.076778, 5.102993, 5.129411, 5.156034, 5.182864, 5.209903,
                                  5.237156, 5.264625, 5.292312, 5.32022, 5.348354, 5.376714, 5.405306, 5.434131, 5.463193,
                          5.492496, 5.522042, 5.551836, 5.58188, 5.612178, 5.642734, 5.673552, 5.704634, 5.735986, 5.76761,
                                  5.799512, 5.831694, 5.864161, 5.896918, 5.929968, 5.963316, 5.996967, 6.030925, 6.065194,
                          6.09978, 6.134687, 6.169921, 6.205486, 6.241387, 6.27763, 6.31422, 6.351163, 6.388465, 6.42613, 6.464166,
                                  6.502578, 6.541371, 6.580553, 6.62013, 6.660109, 6.700495, 6.741297, 6.78252, 6.824173, 6.866262,
                                  6.908795, 6.95178, 6.995225, 7.039137, 7.083525, 7.128398, 7.173764, 7.219632, 7.266011, 7.31291,
                                  7.360339, 7.408308, 7.456827, 7.505905, 7.555554, 7.605785, 7.656608, 7.708035, 7.760077,
                                  7.812747, 7.866057, 7.920019, 7.974647, 8.029953, 8.085952, 8.142657, 8.200083, 8.258245,
                                  8.317158, 8.376837, 8.4373, 8.498562, 8.560641, 8.623554, 8.687319, 8.751955, 8.817481, 8.883916,
                                  8.951282, 9.0196, 9.088889, 9.159174, 9.230477, 9.302822, 9.376233, 9.450735, 9.526355, 9.603118,
                                  9.681054, 9.760191, 9.840558, 9.922186, 10.005107, 10.089353, 10.174959, 10.261958, 10.350389,
                                  10.440287, 10.531693, 10.624646, 10.719188, 10.815362, 10.913214, 11.012789, 11.114137,
                          11.217307, 11.322352, 11.429325, 11.538283, 11.649285, 11.76239, 11.877664, 11.99517, 12.114979,
                                  12.237161, 12.361791, 12.488946, 12.618708, 12.751161, 12.886394, 13.024498, 13.16557, 13.309711,
                                  13.457026, 13.607625, 13.761625, 13.919145, 14.080314, 14.245263, 14.414134, 14.587072,
                          14.764233, 14.945778, 15.131877, 15.322712, 15.51847, 15.719353, 15.92557, 16.137345, 16.354912,
                          16.57852, 16.808433, 17.044929, 17.288305, 17.538873, 17.796967, 18.062943, 18.337176, 18.620068,
                                  18.912049, 19.213574, 19.525133, 19.847249, 20.18048, 20.525429, 20.882738, 21.253102, 21.637266,
                                  22.036036, 22.450278, 22.880933, 23.329017, 23.795634, 24.281981, 24.789364, 25.319207,
                          25.873062, 26.452634, 27.059789, 27.696581, 28.365274, 29.06837, 29.808638, 30.589157, 31.413354,
                                  32.28506, 33.208568, 34.188705, 35.23092, 36.341388, 37.527131, 38.796172, 40.157721, 41.622399,
                                  43.202525, 44.912465, 46.769077, 48.792279, 51.005773, 53.437996, 56.123356, 59.103894 };

double landau(double pMean, double pWidth)
{
    double rc, r;

    /* Local variables */
    int i;
    double u, v;
    double x=randuniform();

    u = x * 1000;
    i = (int) u ;
    u -= (double)i ;
    if (i >= 70 && i <= 800)
    {
        rc = __f[i-6] + u * (__f[i-5]-__f[i-6]);
    }
    else if (i >= 7 && i <= 980)
    {
        rc = __f[i-6] + u * (__f[i-5]-__f[i-6]-(1-u) * .25 * (__f[i-4]-__f[i-5]-__f[i-6] + __f[i-7]));
    }
    else if (i < 7)
    {
        v = log(x);
        u = 1 / v;
        rc = ((u * 17.0854528 + 34.5213058) * u + .9985895) / ((u * 4.01244582 + 34.1760202) * u + 1.) * (-log(-0.91893853-v)-1);
    }
    else
    {
        u = 1-x;
        /* Computing 2nd power */
        r = u;
        v = r * r;
        if (x <= .999)
        {
            rc = (u * 263.991156 + 1.00060006 + v * 4373.20068) / ((u * 257.368075 + 1 + v * 3414.48018) * u);
        }
        else
        {
            rc = (u * 6075.14119 + 1.00001538 + v * 734266.409) / ((u * 6065.11919 + 1 + v * 694021.044) * u);
        }
    }

    return rc*pWidth+pMean;

}
