#include <glib.h>
#ifdef HAVE_CONFIG_H
#ifdef G_OS_WIN32
#include "config_win.h"
#else
#include <config.h>
#endif
#else
#warning "NO config"
#endif

#ifndef HAVE_GIO
#include <iostream>
#include <map>
#include <cstdlib>
#include <cstdio>
#include <cstdarg>
#include <cerrno>
#include <csignal>
#include <cstring>
#include <ctime>
#include <cmath>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/poll.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <glib.h>

#include "Fifo.h"
#include "USBemulator.h"
#include "utils.h"
#include "alibava-server-no-gio.h"


#if GLIB_CHECK_VERSION(2,4,0)
#else
typedef int GPid;
#endif
void __g_gpid_kill( GPid pid)
{
#ifdef G_OS_WIN32
    if (!TerminateProcess(reinterpret_cast<void *>( pid ), 0))
    {
        gchar *msg = g_win32_error_message(GetLastError());
        g_warning("Could not kill child process: %s", msg ? msg : "(null)");
        g_free(msg);
    }
#else /* !G_OS_WIN32 */
    if (kill(pid, SIGKILL))
    {
        g_warning("Could not kill child process: %s", g_strerror(errno));
    }
#endif /* G_OS_WIN32 */
}


/*
 * Called when a sigpipe is received
 */
void sig_pipe(int)
{
    err_msg("*** Lost connection with server ***\n");
    exit(1);
}

/*
 * This function creates a connection to the server.
 */
static int connect_server(char type, const char *name)
{
    int fd, len;
    struct sockaddr_un unix_addr;

    /* Create first a socket that will identify us in the server. */
    if ( (fd=socket(AF_UNIX,SOCK_STREAM,0))<0)
    {
        err_msg("Could not create a socket\n");
        return -1;
    }

    memset(&unix_addr, 0, sizeof(unix_addr));
    unix_addr.sun_family=AF_UNIX;
    sprintf(unix_addr.sun_path, "/tmp/%caps%05d", type, getpid());
    len = strlen(unix_addr.sun_path)+sizeof(unix_addr.sun_family);
    unlink(unix_addr.sun_path);

    if (bind(fd, (struct sockaddr *)&unix_addr, len)<0)
    {
        err_msg("Could not bind client socket\n");
        return -2;
    }
    if (chmod(unix_addr.sun_path, S_IRWXU)<0)
    {
        err_msg("Could not change file permissions\n");
        return -3;
    }

    // Now connect
    memset(&unix_addr, 0, sizeof(unix_addr));
    unix_addr.sun_family=AF_UNIX;
    strcpy(unix_addr.sun_path, name);
    len = strlen(unix_addr.sun_path)+sizeof(unix_addr.sun_family);

    if (connect(fd, (struct sockaddr *)&unix_addr, len)<0)
    {
        err_msg("Could not connect to server\n");
        return -4;
    }
    
    
    int flgs = fcntl(fd, F_GETFL) & ~O_NONBLOCK;
    fcntl(fd, F_SETFL, flgs);
    return fd;
}


static void kill_daemon()
{
    FILE *pf;
    char line[128];

    sprintf(line,"/bin/ps -C %s", DAEMON_NAME);
    pf = popen(line, "r");
    if (!pf)
        return;

    while (fgets(line, sizeof(line), pf))
    {
        if (strstr(line, DAEMON_NAME))
        {
            int pid;
            if ( sscanf(line,"%d",&pid)==1 )
            {
                kill(pid, 9);
            }
        }
    }
    pclose(pf);
}
#if GLIB_CHECK_VERSION(2,4,0)
static void
cb_child_watch( GPid  pid,
                gint  status,
                gpointer data )
{

    /* Close pid */
    g_spawn_close_pid( pid );
}
#endif
static gboolean
cb_out_watch( GIOChannel   *channel,
              GIOCondition  cond,
              gpointer      data )
{
    gchar *string;
    gsize  size;

    if( cond == G_IO_HUP )
    {
        g_io_channel_unref( channel );
        return( FALSE );
    }

    g_io_channel_read_line( channel, &string, &size, NULL, NULL );
    if ( size )
    {
        g_print("[Server] %s", string);
        g_free( string );
        return( TRUE );
    }
    else
        return FALSE;
}

static GPid check_for_daemon(int firmware)
{
    char line[128];
    char *cmmd;
    GError *error = 0;
    GPid pid;
    gint argc;
    gchar **argv;
    gchar *buff;
    gsize buff_len, tpos;
    gint out;
    GIOChannel *out_ch;

    kill_daemon();

    cmmd = getenv("ALIBAVAD");
    if (!cmmd)
    {
        sprintf(line, "%s/%s --daemon=no --firmware-version=%d", DAEMON_PATH, DAEMON_NAME, firmware);
        g_shell_parse_argv(line, &argc, &argv, &error);
        if ( error )
        {
            g_printerr("check_for_daemon: %s\n", error->message);
            g_error_free(error);
            exit(1);
        }
    }
    else
    {
        g_shell_parse_argv(cmmd, &argc, &argv, &error);
        if ( error )
        {
            g_printerr("check_for_daemon (ALIBAVAD): %s\n", error->message);
            g_error_free(error);
            exit(1);
        }
    }
    /*
     * TODO: which is the workding directory of alibavad ?
     */
    if  (! g_spawn_async_with_pipes(NULL, argv, NULL,
                                    G_SPAWN_DO_NOT_REAP_CHILD,
                                    NULL, NULL,
                                    &pid, NULL, NULL,&out, &error) )
    {
        g_printerr("Could not start daemon: %s\n", error->message);
        g_error_free(error);
        exit(1);
    }
#if GLIB_CHECK_VERSION(2,4,0)
    g_child_watch_add( pid, (GChildWatchFunc)cb_child_watch, NULL );
#endif
#ifdef G_OS_WIN32
    out_ch = g_io_channel_win32_new_fd( out );
#else
    out_ch = g_io_channel_unix_new( out );
#endif

    error = 0;
    buff_len = tpos = 0;
    g_io_channel_read_line(out_ch, &buff, &buff_len, &tpos, &error);
    if ( error )
    {
        g_printerr("Could not start daemon: %s\n", error->message);
        g_error_free(error);
        exit(1);
    }
    g_free(buff);

    g_io_add_watch( out_ch, (GIOCondition)(G_IO_IN | G_IO_HUP), (GIOFunc)cb_out_watch, 0 );
    return pid;
}

//GData *fildes;
std::map<int, int> file_map;

/*
 * We connect to a daemon twice per process. One is for the
 * ioctl/lseek operations and the other one is for read/write. We fork
 * a subprocess that will behave as the kernel module.
 */
int do_open(const char *path, int flags, int firmware, GPid *_pid)
{
    int fd=-1, cfd=-1, flgs;
    GPid pid;
	unsigned long handshake;
    GError *error = 0;
    file_map.clear();

    pid = check_for_daemon(firmware);
    if (_pid)
        *_pid = pid;

    // Create the ioctl channel
    int mx_try = 10;
    for (int itry=0; itry<mx_try && cfd<0; itry++)
    {
       cfd = connect_server('c', SOCKET_NAME);
       if (cfd>=0)
          break;
          
        g_usleep(1000);
     }
     if (cfd<0)
        return -1;
     else
     {
        int opt_val = 1;
        setsockopt( cfd,
                    IPPROTO_TCP,
                    TCP_NODELAY, (char *)&opt_val,
                    sizeof(int));
        int rc = read(cfd, &handshake, sizeof(unsigned long));
        if (rc<0)
        {
         err_msg(_("Cannot receive handshake word: %s\n"), strerror(errno) );
        }
     }

    // Create the I/O pipe

    fd = connect_server('m', SOCKET_NAME);
    if (fd<0)
        return -1;
    else
    {
    	int opt_val = 1;
        setsockopt( fd,
                    IPPROTO_TCP,
                    TCP_NODELAY, (char *)&opt_val,
                    sizeof(int));

        int rc = read(fd, &handshake, sizeof(unsigned long));
        if (rc<0)
        {
         err_msg(_("Cannot receive handshake word: %s\n"), strerror(errno) );
        }
    }


    file_map[fd] = cfd;

    signal(SIGPIPE, sig_pipe);

    return fd;
}

/*
 *---------------------------------------------------------------
 * The client routines.
 * These routines are called by the client to comunicate with the
 * process that fakes the driver.
 *---------------------------------------------------------------
 */



/*
 * Send the ioctl commands to the "driver"
 */
int do_ioctl(int fd, int req, ...)
{
    va_list args;
    void *ptr;
    int rc, tmp;
    long lval;
    int dir= _IOC_DIR (req);
    int sz= _IOC_SIZE (req);

    int cfd = file_map[fd];

    PDEBUG ("sending %x\n", req);

    /*
     * If we want the server to DIE, we do not make any other
     * complicated stuff.
     */
    if (req == DIE)
    {
        PDEBUG ("This is the end\n");
        rc = write(cfd, &req, sizeof(int));
        rc = read(cfd, &tmp, sizeof(int));
        return 0;
    }

    // Otherwise, send the request
    rc = write(cfd, &req, sizeof(int));
    if (rc < 0)
        return rc;

    // See what else is needed. We can set get or both
    if (dir & (_IOC_READ | _IOC_WRITE))
    {
        va_start(args, req);
        lval = va_arg (args, long);
        ptr = (char *)&lval;

        if (dir & _IOC_WRITE)
        {
            // we have to send a value
            write(cfd, ptr, sz);
        }
        if (dir & _IOC_READ)
        {
            int rc;
            // we have to read something back
            ptr = (char *) lval;
            rc = read(cfd, ptr, sz);
            if (rc<=0)
            {
                err_msg("problems reading back the value");
            }
        }
        va_end(args);
    }
    // get the return code;
    rc = read(cfd, &tmp, sizeof(int));

    return tmp;
}

/*
 * Reading from the server.
 */
int do_read(int fd, void *buff, size_t size)
{
    int iok, rc;
    rc = do_ioctl(fd, DO_READ, size);
    iok = read(fd, buff, size);
    rc = do_ioctl(fd, ERROR);
    if (rc<0 || iok<=0)
        return -1;

    return iok;
}

/*
 * Write is just a normal write. The server should be able to retrieve
 * the ammount of information with FIONREAD
 */
int do_write(int fd, void *buff, size_t size)
{
    int rc = write(fd, buff, size);
    if ( rc<0 )
    {
        err_msg(_("Error writing to socket: (%d) %s\n"), rc, strerror(errno));
    }
    return rc;
}

/*
 * The close operation. Send the DIE message to the daemon and
 * close
 */
int do_close(int fd)
{
    int rc;
    int cfd = file_map[fd];
    rc = do_ioctl(fd, DIE);
    g_usleep(100000);
    close(fd);
    close(cfd);

    file_map.erase(fd);
    return 0;
}
struct USBemulator_priv
{
        int S;
        GIOChannel *chan;
        GPid pid;
        bool use_fifo;
};

USBemulator::USBemulator(int firmware)
    : _firmware(firmware), _priv(0)
{
    _priv = g_new0(USBemulator_priv, 1);
    _priv->S = -1;
    _priv->chan = 0;
    _priv->pid = 0;
    _priv->use_fifo = false;
    fd = 0;
}

USBemulator::~USBemulator()
{
    if (_priv->S >=0)
        close();

    kill_daemon();
    __g_gpid_kill(_priv->pid);
#if GLIB_CHECK_VERSION(2,4,0)
    g_spawn_close_pid (_priv->pid);
#endif
    g_free(_priv);
}

int USBemulator::error() const
{

    return _priv->S ? do_ioctl(_priv->S, ERROR_NR) : -1;
}

int USBemulator::open(const std::string &fnam)
{
    fd = _priv->S = do_open(fnam.c_str(), O_RDWR, _firmware, &_priv->pid);
    if (_priv->S<0)
    {
        err_msg(_("do_open does not return a valid socket"));
        return -1;
    }
    else
    {
        GError *error = 0;
#ifdef G_OS_WIN32
      _priv->chan = g_io_channel_win32_new_socket( g_socket_get_fd(_priv->S));
#else
      _priv->chan = g_io_channel_unix_new( _priv->S );
#endif
      g_io_channel_set_encoding(_priv->chan, 0, &error);
      g_io_channel_set_buffered(_priv->chan, FALSE);
      g_io_channel_ref(_priv->chan);
    }
    return 0;
}

int USBemulator::close()
{
    if ( _priv->S>=0)
    {

        return do_close(_priv->S);
    }
    else
        return 0;
}

int USBemulator::read(void *buf, int size)
{
    return do_read(_priv->S, buf, size);
}

int USBemulator::write(void *buf, int size)
{
    return do_write(_priv->S, buf, size);
}

int USBemulator::poll(int timeout)
{
    return USBserial::poll(timeout);
}

int USBemulator::setup_fifo(Fifo *f)
{
    _priv->use_fifo = true;
    f->attach(_priv->S);

    return 0;
}
#endif
