/*
 * mingw.c
 *
 *  Created on: 20/11/2011
 *      Author: Carlos
 */
#include "mingw.h"
#include <cstdio>

#ifdef G_OS_WIN32


static int __ws_initialized__ = 0;
static void __init_bloody_ws__()
{
    if ( ! __ws_initialized__ )
    {
        WSADATA wsaData;   // if this doesn't work
        //WSAData wsaData; // then try this instead

        // MAKEWORD(1,1) for Winsock 1.1, MAKEWORD(2,0) for Winsock 2.0:


        if (WSAStartup(MAKEWORD(2,2), &wsaData) != 0) {
            fprintf(stderr, "WSAStartup failed.\n");
            exit(1);
        }
    }
    __ws_initialized__ ++;
}
static void __cleanup_bloody_ws__()
{
    __ws_initialized__ --;
    if (!__ws_initialized__)
        WSACleanup();
}

class WSocketInitializer
{
    public:
        WSocketInitializer()
        {
            __init_bloody_ws__();
        }
        ~WSocketInitializer()
        {
            __cleanup_bloody_ws__();
        }
};
static WSocketInitializer __ws_initializer__;

int poll(struct pollfd *fds, unsigned int nfds, int timo)
{
    struct timeval timeout, *toptr;
    fd_set ifds, ofds, efds, *ip, *op;
    int i, rc;

    /* Set up the file-descriptor sets in ifds, ofds and efds. */
    FD_ZERO(&ifds);
    FD_ZERO(&ofds);
    FD_ZERO(&efds);
    for (i = 0, op = ip = 0; i < nfds; ++i) {
    fds[i].revents = 0;
    if(fds[i].events & (POLLIN|POLLPRI)) {
        ip = &ifds;
        FD_SET(fds[i].fd, ip);
    }
    if(fds[i].events & POLLOUT) {
        op = &ofds;
        FD_SET(fds[i].fd, op);
    }
    FD_SET(fds[i].fd, &efds);
    }

    /* Set up the timeval structure for the timeout parameter */
    if(timo < 0) {
    toptr = 0;
    } else {
    toptr = &timeout;
    timeout.tv_sec = timo / 1000;
    timeout.tv_usec = (timo - timeout.tv_sec * 1000) * 1000;
    }

#ifdef DEBUG_POLL
    printf("Entering select() sec=%ld usec=%ld ip=%lx op=%lx\n",
           (long)timeout.tv_sec, (long)timeout.tv_usec, (long)ip, (long)op);
#endif
    rc = select(0, ip, op, &efds, toptr);
#ifdef DEBUG_POLL
    printf("Exiting select rc=%d\n", rc);
#endif

    if(rc <= 0)
    return rc;

    if(rc > 0) {
        for (i = 0; i < nfds; ++i) {
            int fd = fds[i].fd;
        if(fds[i].events & (POLLIN|POLLPRI) && FD_ISSET(fd, &ifds))
            fds[i].revents |= POLLIN;
        if(fds[i].events & POLLOUT && FD_ISSET(fd, &ofds))
            fds[i].revents |= POLLOUT;
        if(FD_ISSET(fd, &efds))
            /* Some error was detected ... should be some way to know. */
            fds[i].revents |= POLLHUP;
#ifdef DEBUG_POLL
        printf("%d %d %d revent = %x\n",
                FD_ISSET(fd, &ifds), FD_ISSET(fd, &ofds), FD_ISSET(fd, &efds),
                fds[i].revents
        );
#endif
        }
    }
    return rc;
}


#endif



#include <gtk/gtk.h>

#ifdef G_OS_WIN32
static char appfontname[128] = "tahoma 8"; /* fallback value */
#else
static char appfontname[128] = "Sans 10";
#endif

void set_app_font(const char *fontname) {
	GtkSettings *settings;

	if (fontname != NULL && *fontname == 0)
		return;

	settings = gtk_settings_get_default();

	if (fontname == NULL)
	{
		g_object_set(G_OBJECT(settings), "gtk-font-name", appfontname, NULL);
	}
	else
	{
		GtkWidget *w;
		PangoFontDescription *pfd;
		PangoContext *pc;
		PangoFont *pfont;

		w = gtk_label_new(NULL);
		pfd = pango_font_description_from_string(fontname);
		pc = gtk_widget_get_pango_context(w);
		pfont = pango_context_load_font(pc, pfd);

		if (pfont != NULL)
		{
			strcpy(appfontname, fontname);
			g_object_set(G_OBJECT(settings), "gtk-font-name", appfontname, NULL);
		}

		gtk_widget_destroy(w);
		pango_font_description_free(pfd);
	}
}

#ifdef G_OS_WIN32
char *default_windows_menu_fontspec(void)
{
    gchar *fontspec = NULL;
    NONCLIENTMETRICS ncm;

    memset(&ncm, 0, sizeof ncm);
    ncm.cbSize = sizeof ncm;

    if (SystemParametersInfo(SPI_GETNONCLIENTMETRICS, ncm.cbSize, &ncm, 0))
    {
        HDC screen = GetDC(0);
        double y_scale = 72.0 / GetDeviceCaps(screen, LOGPIXELSY);
        int point_size = (int) (ncm.lfMenuFont.lfHeight * y_scale);

        if (point_size < 0)
            point_size = -point_size;
        fontspec = g_strdup_printf("%s %d", ncm.lfMenuFont.lfFaceName, point_size);
        ReleaseDC(0, screen);
    }

    return fontspec;
}

void try_to_get_windows_font(void)
{
    gchar *fontspec = default_windows_menu_fontspec();

    if (fontspec != NULL)
    {
        int match = 0;
        PangoFontDescription *pfd;
        PangoFont *pfont;
        PangoContext *pc;
        GtkWidget *w;

        pfd = pango_font_description_from_string(fontspec);

        w = gtk_label_new(NULL);
        pc = gtk_widget_get_pango_context(w);
        pfont = pango_context_load_font(pc, pfd);
        match = (pfont != NULL);

        pango_font_description_free(pfd);
        g_object_unref(G_OBJECT(pc));
        gtk_widget_destroy(w);

        if (match)
            set_app_font(fontspec);
        g_free(fontspec);
    }
}


#define TOTALBYTES    8192
#define BYTEINCREMENT 4096
gchar *get_registry_key(const gchar *path, const gchar *key)
{
    DWORD BufferSize = TOTALBYTES;
    DWORD cbData;
    DWORD dwRet = ERROR_SUCCESS;
    HKEY hKey;
    PPERF_DATA_BLOCK PerfData = 0;

    if( RegOpenKeyEx(HKEY_LOCAL_MACHINE, path,0, KEY_QUERY_VALUE, &hKey) == ERROR_SUCCESS)
    {
        PerfData = (PPERF_DATA_BLOCK) g_malloc( BufferSize );
        cbData = BufferSize;

        dwRet = RegQueryValueEx( hKey,
                                 key,
                                 NULL,
                                 NULL,
                                 (LPBYTE) PerfData,
                                 &cbData );
        while( dwRet == ERROR_MORE_DATA )
        {
            // Get a buffer that is big enough.

            BufferSize += BYTEINCREMENT;
            PerfData = (PPERF_DATA_BLOCK) g_realloc( PerfData, BufferSize );
            cbData = BufferSize;

            dwRet = RegQueryValueEx( hKey,
                                     key,
                                     NULL,
                                     NULL,
                                     (LPBYTE) PerfData,
                                     &cbData );
        }
        if( dwRet != ERROR_SUCCESS )
        {
            free( (void *)PerfData );
            RegCloseKey(hKey);
            return 0;
        }


    }
    RegCloseKey(hKey);

    return (gchar *)PerfData;
}



//==================================
// This is a modification of:
//   PEDUMP - Matt Pietrek 1995
// done by Ismael Jurado 1997
//
// This program only outputs to stdout a DEF file from a DLL given in the
// command line.
// I don´t know how to contact Matt, so if this program is not legal, please
// let me know.
//    ismaelj@hotmail.com
//==================================

#include <windows.h>
#include <stdio.h>


#define _WINNT_

// This code was taken from WINNT.H, which is not included in the current
// Cygnus port.
#ifndef _WINNT_
#define FIELD_OFFSET(type, field)    ((LONG)&(((type *)0)->field))
#define IMAGE_FIRST_SECTION( ntheader ) ((PIMAGE_SECTION_HEADER)        \
    ((DWORD)ntheader +                                                  \
     FIELD_OFFSET( IMAGE_NT_HEADERS, OptionalHeader ) +                 \
     ((PIMAGE_NT_HEADERS)(ntheader))->FileHeader.SizeOfOptionalHeader   \
    ))

#define IMAGE_SIZEOF_SHORT_NAME              8
#define IMAGE_DIRECTORY_ENTRY_EXPORT         0   // Export Directory
#define IMAGE_NUMBEROF_DIRECTORY_ENTRIES    16
#define IMAGE_NT_SIGNATURE                  0x00004550  // PE00
#define IMAGE_DOS_SIGNATURE                 0x5A4D      // MZ
typedef struct _IMAGE_DATA_DIRECTORY
{
    DWORD VirtualAddress;
    DWORD Size;
}IMAGE_DATA_DIRECTORY, *PIMAGE_DATA_DIRECTORY;

typedef struct _IMAGE_FILE_HEADER
{
    WORD Machine;
    WORD NumberOfSections;
    DWORD TimeDateStamp;
    DWORD PointerToSymbolTable;
    DWORD NumberOfSymbols;
    WORD SizeOfOptionalHeader;
    WORD Characteristics;
}IMAGE_FILE_HEADER, *PIMAGE_FILE_HEADER;

typedef struct _IMAGE_OPTIONAL_HEADER
{
    //
    // Standard fields.
    //

    WORD Magic;
    BYTE MajorLinkerVersion;
    BYTE MinorLinkerVersion;
    DWORD SizeOfCode;
    DWORD SizeOfInitializedData;
    DWORD SizeOfUninitializedData;
    DWORD AddressOfEntryPoint;
    DWORD BaseOfCode;
    DWORD BaseOfData;

    //
    // NT additional fields.
    //

    DWORD ImageBase;
    DWORD SectionAlignment;
    DWORD FileAlignment;
    WORD MajorOperatingSystemVersion;
    WORD MinorOperatingSystemVersion;
    WORD MajorImageVersion;
    WORD MinorImageVersion;
    WORD MajorSubsystemVersion;
    WORD MinorSubsystemVersion;
    DWORD Reserved1;
    DWORD SizeOfImage;
    DWORD SizeOfHeaders;
    DWORD CheckSum;
    WORD Subsystem;
    WORD DllCharacteristics;
    DWORD SizeOfStackReserve;
    DWORD SizeOfStackCommit;
    DWORD SizeOfHeapReserve;
    DWORD SizeOfHeapCommit;
    DWORD LoaderFlags;
    DWORD NumberOfRvaAndSizes;
    IMAGE_DATA_DIRECTORY DataDirectory[IMAGE_NUMBEROF_DIRECTORY_ENTRIES];
}IMAGE_OPTIONAL_HEADER, *PIMAGE_OPTIONAL_HEADER;

typedef struct _IMAGE_NT_HEADERS
{
    DWORD Signature;
    IMAGE_FILE_HEADER FileHeader;
    IMAGE_OPTIONAL_HEADER OptionalHeader;
}IMAGE_NT_HEADERS, *PIMAGE_NT_HEADERS;

typedef struct _IMAGE_SECTION_HEADER
{
    BYTE Name[IMAGE_SIZEOF_SHORT_NAME];
    union
    {
        DWORD PhysicalAddress;
        DWORD VirtualSize;
    }Misc;
    DWORD VirtualAddress;
    DWORD SizeOfRawData;
    DWORD PointerToRawData;
    DWORD PointerToRelocations;
    DWORD PointerToLinenumbers;
    WORD NumberOfRelocations;
    WORD NumberOfLinenumbers;
    DWORD Characteristics;
}IMAGE_SECTION_HEADER, *PIMAGE_SECTION_HEADER;

typedef struct _IMAGE_EXPORT_DIRECTORY
{
    DWORD Characteristics;
    DWORD TimeDateStamp;
    WORD MajorVersion;
    WORD MinorVersion;
    DWORD Name;
    DWORD Base;
    DWORD NumberOfFunctions;
    DWORD NumberOfNames;
    PDWORD *AddressOfFunctions;
    PDWORD *AddressOfNames;
    PWORD *AddressOfNameOrdinals;
}IMAGE_EXPORT_DIRECTORY, *PIMAGE_EXPORT_DIRECTORY;

typedef struct _IMAGE_DOS_HEADER
{ // DOS .EXE header
    WORD e_magic;// Magic number
    WORD e_cblp;// Bytes on last page of file
    WORD e_cp;// Pages in file
    WORD e_crlc;// Relocations
    WORD e_cparhdr;// Size of header in paragraphs
    WORD e_minalloc;// Minimum extra paragraphs needed
    WORD e_maxalloc;// Maximum extra paragraphs needed
    WORD e_ss;// Initial (relative) SS value
    WORD e_sp;// Initial SP value
    WORD e_csum;// Checksum
    WORD e_ip;// Initial IP value
    WORD e_cs;// Initial (relative) CS value
    WORD e_lfarlc;// File address of relocation table
    WORD e_ovno;// Overlay number
    WORD e_res[4];// Reserved words
    WORD e_oemid;// OEM identifier (for e_oeminfo)
    WORD e_oeminfo;// OEM information; e_oemid specific
    WORD e_res2[10];// Reserved words
    LONG e_lfanew;// File address of new exe header
}IMAGE_DOS_HEADER, *PIMAGE_DOS_HEADER;
#endif // __WINNT__
// Here starts Matt's code
//
//
// MakePtr is a macro that allows you to easily add to values (including
// pointers) together without dealing with C's pointer arithmetic.  It
// essentially treats the last two parameters as DWORDs.  The first
// parameter is used to typecast the result to the appropriate pointer type.
#define MakePtr( cast, ptr, addValue ) (cast)( (DWORD)(ptr) + (DWORD)(addValue))
PIMAGE_SECTION_HEADER GetEnclosingSectionHeader(DWORD rva, PIMAGE_NT_HEADERS pNTHeader)
{
    PIMAGE_SECTION_HEADER section = IMAGE_FIRST_SECTION(pNTHeader);
    unsigned i;

    for (i = 0; i < pNTHeader->FileHeader.NumberOfSections; i++, section++)
    {
        // Is the RVA within this section?
        if ((rva >= section->VirtualAddress)
                && (rva < (section->VirtualAddress + section->Misc.VirtualSize)))
            return section;
    }
    return 0;
}
#include <vector>
#include <string>
typedef std::vector<std::string> SymbolList;

void DumpExportsSection(DWORD base, PIMAGE_NT_HEADERS pNTHeader, SymbolList &s_list)
{
    PIMAGE_EXPORT_DIRECTORY exportDir;
    PIMAGE_SECTION_HEADER header;
    INT delta;
    DWORD i;
    PDWORD functions;
    PWORD ordinals;
    PSTR *name;
    DWORD exportsStartRVA, exportsEndRVA;

    exportsStartRVA =
            pNTHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].VirtualAddress;
    exportsEndRVA = exportsStartRVA
            + pNTHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].Size;

    // Get the IMAGE_SECTION_HEADER that contains the exports.  This is
    // usually the .edata section, but doesn't have to be.
    header = (PIMAGE_SECTION_HEADER) GetEnclosingSectionHeader(exportsStartRVA,
                                                               pNTHeader);
    if (!header)
        return;
    delta = (INT) (header->VirtualAddress - header->PointerToRawData);
    exportDir = MakePtr(PIMAGE_EXPORT_DIRECTORY, base, exportsStartRVA - delta);
    //printf("EXPORTS:\n");
    functions = (PDWORD) ((DWORD) exportDir->AddressOfFunctions - delta + base);
    ordinals = (PWORD) ((DWORD) exportDir->AddressOfNameOrdinals - delta + base);
    name = (PSTR *) ((DWORD) exportDir->AddressOfNames - delta + base);
    for (i = 0; i < exportDir->NumberOfFunctions; i++)
    {
        DWORD entryPointRVA = functions[i];
        DWORD j;
        int flag = 0;

        if (entryPointRVA == 0) // Skip over gaps in exported function
            continue; // ordinals (the entrypoint is 0 for
                      // these functions).
        // See if this function has an associated name exported for it.
        for (j = 0; j < exportDir->NumberOfNames; j++)
            if (ordinals[j] == i)
            {
                char __line__[256];
                sprintf(__line__,"%s", name[j] - delta + base);
                s_list.push_back( __line__ );
            }
        // Is it a forwarder?  If so, the entry point RVA is inside the
        // .edata section, and is an RVA to the DllName.EntryPointName
        if ((entryPointRVA >= exportsStartRVA) && (entryPointRVA <= exportsEndRVA))
        {
            char __line__[256];
            sprintf(__line__, "%s", entryPointRVA - delta + base);
            s_list.push_back( __line__ );
        }
    }
}

int DumpExeFile(PIMAGE_DOS_HEADER dosHeader, SymbolList &s_list)
{
    PIMAGE_NT_HEADERS pNTHeader;
    DWORD base = (DWORD) dosHeader;

    pNTHeader = MakePtr( PIMAGE_NT_HEADERS, dosHeader,
            dosHeader->e_lfanew );

    // First, verify that the e_lfanew field gave us a reasonable
    // pointer, then verify the PE signature.
    // IJ: gcc doesn't support C exeptions??
//    __try
//    {
    if (pNTHeader->Signature != IMAGE_NT_SIGNATURE)
    {
        printf("Not a Portable Executable (PE) EXE\n");
        return 1;
    }
//    }
    /*
     __except( TRUE )    // Should only get here if pNTHeader (above) is bogus
     {
     printf( "invalid .EXE\n");
     return;
     }
     */
    DumpExportsSection(base, pNTHeader, s_list);
    //printf("\n");
    return 0;
}

int DumpFile(LPSTR filename, SymbolList &s_list)
{
    HANDLE hFile;
    HANDLE hFileMapping;
    LPVOID lpFileBase;
    PIMAGE_DOS_HEADER dosHeader;

    hFile = CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,
                       FILE_ATTRIBUTE_NORMAL, 0);

    if (hFile == INVALID_HANDLE_VALUE)
    {
        printf("Couldn't open file with CreateFile()\n");
        return 1;
    }

    hFileMapping = CreateFileMapping(hFile, NULL, PAGE_READONLY, 0, 0, NULL);
    if (hFileMapping == 0)
    {
        CloseHandle(hFile);
        printf("Couldn't open file mapping with CreateFileMapping()\n");
        return 1;
    }

    lpFileBase = MapViewOfFile(hFileMapping, FILE_MAP_READ, 0, 0, 0);
    if (lpFileBase == 0)
    {
        CloseHandle(hFileMapping);
        CloseHandle(hFile);
        printf("Couldn't map view of file with MapViewOfFile()\n");
        return 1;
    }

    //printf(";Exports of file %s\n\n", filename);

    dosHeader = (PIMAGE_DOS_HEADER) lpFileBase;
    if (dosHeader->e_magic == IMAGE_DOS_SIGNATURE)
    {
        return DumpExeFile(dosHeader, s_list);
    }
    else
    {
        printf("Format Unknown!\n");
        return 1;
    }
}

/*

static char *dlerr_ptr;
static char dlerr_data[80];

void *dlopen (const char *file, int mode)
{
   HMODULE hmodule;

   hmodule = LoadLibrary(file);
   if (hmodule == NULL) {
     int error;

     error = GetLastError();
     sprintf(dlerr_data, "LoadLibraryEx returned %d.", error);
     dlerr_ptr = dlerr_data;
   }
   return hmodule;
}

int dlclose (void *handle)
{
   if (FreeLibrary(handle)) {
     return 0;
   }
   else {
     int error;

     error = GetLastError();
     sprintf(dlerr_data, "FreeLibrary returned %d.", error);
     dlerr_ptr = dlerr_data;
     return -1;
   }
}

void *dlsym (void *handle, const char *name)
{
   FARPROC fp;

   fp = GetProcAddress(handle, name);
   if (fp == NULL) {
     int error;

     error = GetLastError();
     sprintf(dlerr_data, "GetProcAddress returned %d.", error);
     dlerr_ptr = dlerr_data;
   }
   return fp;
}

char *dlerror (void)
{
   if (dlerr_ptr != NULL) {
     dlerr_ptr = NULL;
     return dlerr_data;
   }
   else {
     return NULL;
   }
}

*/


#endif /* G_OS_WIN32 */



