#!/usr/bin/env python
""" 
    Calibrate the clock counter.
"""
import sys
from pylab import *
import h5py

def main(fnam):
    """ Shows the main stuff
    """
    figure(figsize=(7.5,10))
#    subplots_adjust(hspace=0.5, top=0.975, bottom=0.025)


    F = h5py.File(fnam, 'r')


    # Get the clock counter
    C = F['/events/clock']
    T = F['/events/time']

    # And the time differents between consecutive events
    CC0  = (C[1:] - C[:-1])
    CC1  = CC0.view('float32')
    CC1[:] = CC0
    
    Ta = T[:].view('float32')
    Ta[:] = T

    indx = nonzero( abs(CC1-CC1[0])<3000. )
    nindx = nonzero( abs(CC1-CC1[0])>3000. )
    CC  = CC1[ indx ]*25.0e-3
    DD  = CC1[ nindx ]*25.0e-9
    TT  = Ta[ indx ]
    print "N. evts", CC.shape
    
    # Get the mean value and std deviation
    avg = mean(CC)
    rms = std(CC)
    print "mean %f usec / rms %f usec" %(avg, rms)

    x0 = avg-5*rms
    x1 = avg+5*rms
    stp = (x1-x0)/1000.0
    X = r_[x0:x1:stp]
    n, bins, patches = hist(CC, X)

    print CC[250000:250100]
    print DD[1:10]

    nplot = 4

    subplot(nplot, 1, 1)
    plot(CC)
    title('Clock counter (usec) .vs. event no.')
    xlabel('Event Number')

    subplot(nplot, 1, 2)
    #plot(DD)
    nd, dbins, dpatches = hist(DD, 50, normed=1, histtype='stepfilled')
    setp(dpatches, 'facecolor', 'g', 'alpha', 0.75)
    title('Gap between samples (s)')
    xlabel('Sample Number')

    subplot(nplot, 1, 3)
    plot(TT[:100])
    title('TDC time .vs. event no.')
    xlabel('Event Number')

    subplot(nplot, 1, 4)
    plot(bins[:-1], n)
    print n
    title('Mean clock counter value (usec)')
    xlabel('Event Interval (usec)')

    tight_layout()
    print "Number of events", CC.shape[0]
    show()


if __name__ == "__main__":
    try:
        main(sys.argv[1])
    except KeyError:
        print "I need an input file"
        sys.exit(1)
             
